"""Facade classes for working with Shiny inputs/outputs in Playwright"""

# TODO-barret; Possibly move all navset's loc_containers to the parent element (e.g. NavsetCardUnderline should have the container be the containing card.) The `.loc` will then point to the `ul` that has the id
# TODO-barret; This should be the container and the `ul#{id}.navbar-nav` should be the loc
# TODO-barret; Maybe add a `loc_nav_item` that contains `> li.nav-item`?
# TODO-barret; Note: We have access to the panel via `.nav_panel("key")`
# TODO-barret; Maybe add `.loc_sidebar` for the sidebar?


from __future__ import annotations

import typing
from typing import Literal, Protocol

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

# Import `shiny`'s typing extentions.
# Since this is a private file, tell pyright to ignore the import
from ..._typing_extensions import TypeGuard
from ...types import MISSING_TYPE
from .._types import AttrValue, OptionalFloat, PatternOrStr, Timeout
from ..expect._internal import (
    expect_attribute_to_have_value as _expect_attribute_to_have_value,
)
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value

"""
Questions:
* `_DateBase` is signaled as private, but `InputDateRange` will have two fields of `date_start` and `date_end`. Due to how the init selectors are created, they are not `InputDate` instances. Should we make `_DateBase` public?

* While `expect_*_to_have_value()` matches the setup of `expect(x).to_have_value()`, it is a bit verbose. Should we just use `expect_*()` as we only use it in a single context? (Only adding the suffix if other methods like `to_have_html()` or `to_have_text()` would make sense.)
    * Ans: Try things out
    * Ans 3-7-2023: Use small names as there is no need to differentiate using longer names.

* TODO-future; Make sure multiple usage of `timeout` has the proper values. Should followup usages be `0` to force it to be immediate? (Is `0` the right value?)
    * Ans: There was no definition of "now" for a timeout. 0 disables the timeout.
"""

"""
# Class definitions
* Fields
  * Try to mirror playwright as much as possible.
    * Do not use verbose methods as we do not need the long name of
      `expect_label_to_have_text()` as there is not `expect_label_to_have_html()`
      or `expect_label_to_have_attribute()` methods. Just use `expect_label()`
  * There are no properties, only methods; This allows for timeout values to be passed through and for complex methods.
    * Locators will stay as properties
  * Don't sub-class. For now, use `_` separatation and use `loc` or `value` as a prefix
* Approach
  * Use locators / playwright_expect as much as possible
    * It should not be necessary to use `assert` directly.
    * MUST wait for `Locator`s to do their job
  * DO NOT provide `value` methods
  * Add _set_ methods (or set like methods) only if a user would perform them
"""

"""
# Mixins
* Use mixins to add consistent functionality to different classes
* These classes should **never** be instantiated directly
* Use `typing.Protocol` to define the interface of what is required on `self`
* Add methods to the mixin if they are consistently used across multiple classes
  * If a method is only used in one class, it should be defined in that class
  * If a method is used inconsistently, make/use a helper method
"""

InitLocator = typing.Union[Locator, str]

R = typing.TypeVar("R")
M1 = typing.TypeVar("M1")
M2 = typing.TypeVar("M2")


def is_missing(x: object) -> TypeGuard[MISSING_TYPE]:
    return isinstance(x, MISSING_TYPE)


# TypeGuard does not work for `not isinstance(x, MISSING_TYPE)`
# See discussion for `StrictTypeGuard`: https://github.com/python/typing/discussions/1013
# Until then, we need `not_is_missing(x=)` to narrow within an `if` statement
def not_is_missing(x: R | MISSING_TYPE) -> TypeGuard[R]:
    return not isinstance(x, MISSING_TYPE)


def all_missing(*args: object) -> TypeGuard[MISSING_TYPE]:
    for arg in args:
        if not_is_missing(arg):
            return False
    return True


def maybe_missing(x: M1 | MISSING_TYPE, default: M2) -> M1 | M2:
    if isinstance(x, MISSING_TYPE):
        return default
    return x


def set_text(
    loc: Locator,
    text: str,
    *,
    delay: OptionalFloat = None,
    timeout: Timeout = None,
) -> None:
    """
    Sets the text of a DOM element.

    Parameters
    ----------
    loc
        Playwright `Locator` of the element.
    text
        The text to set.
    delay
        The delay between key presses in milliseconds. Defaults to `None`.
    timeout
        The maximum time to wait for the text to be set. Defaults to `None`.
    """
    # TODO-future; Composable set() method
    loc.fill("", timeout=timeout)  # Reset the value
    loc.type(text, delay=delay, timeout=timeout)  # Type the value


def _expect_multiple(loc: Locator, multiple: bool, timeout: Timeout = None) -> None:
    value = "True" if multiple else None
    _expect_style_to_have_value(loc, "multiple", value, timeout=timeout)


######################################################
# # Inputs
######################################################


class UiBaseP(Protocol):
    id: str
    loc: Locator
    page: Page


class UiWithContainerP(UiBaseP, Protocol):
    """A protocol class representing UI with a container."""

    loc_container: Locator
    """
    Playwright `Locator` for the container of the UI element.
    """


class UiWithSidebarP(UiWithContainerP, Protocol):
    """A protocol class representing UI with an associated sidebar."""

    loc_sidebar: Locator
    """
    Playwright `Locator` for its sidebar of the UI element.
    """


class UiWithTitleP(UiWithContainerP, Protocol):
    """A protocol class representing UI with an associated title."""

    loc_title: Locator
    """
    Playwright `Locator` for its title of the UI element.
    """


class UiBase:
    """A base class representing shiny UI components."""

    # timeout: Timeout
    id: str
    """
    The browser DOM `id` of the UI element.
    """
    loc: Locator
    """
    Playwright `Locator` of the UI element.
    """
    page: Page
    """
    Playwright `Page` of the Shiny app.
    """

    def __init__(
        self,
        page: Page,
        *,
        id: str,
        loc: InitLocator,
    ) -> None:
        self.page = page
        # Needed?!? This is covered by `self.loc_root` and possibly `self.loc`
        self.id = id
        if isinstance(loc, str):
            loc = page.locator(loc)
        self.loc = loc

    @property
    # TODO; Can not publicly find `LocatorAssertions` in `playwright`
    def expect(self):
        """Expectation method equivalent to `playwright.expect(self.loc)`."""
        # TODO-karan-test: Search for `.loc)` and convert `expect(FOO.loc)` to `FOO.expect`. If we don't like the helper API, we should remove it.
        return playwright_expect(self.loc)


class UiWithContainer(UiBase):
    """
    A mixin class representing UI with a container.
    """

    loc_container: Locator
    """
    Playwright `Locator` for the container of the UI element.
    """

    def __init__(
        self,
        page: Page,
        *,
        id: str,
        loc: InitLocator,
        loc_container: InitLocator = "div.shiny-input-container",
    ) -> None:
        """
        Initializes the input with a container.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The id of the UI element.
        loc
            Playwright `Locator` of the UI element.
        loc_container
            Playwright `Locator` of the container of the UI element.
        """
        loc_is_str = isinstance(loc, str)
        loc_container_is_str = isinstance(loc_container, str)

        if loc_is_str and loc_container_is_str:
            loc_container = page.locator(loc_container)
            if loc == "xpath=.":
                # If `loc` is self, then use `loc_container` as `loc`
                loc = loc_container

            else:
                loc_container = loc_container.locator(
                    # `page.locator(loc)` is executed from within `loc_container`
                    "xpath=.",
                    has=page.locator(loc),
                )

                loc = loc_container.locator(loc)
        elif not loc_is_str and not loc_container_is_str:
            ...  # Do nothing; Use values as is
        else:
            raise ValueError(
                "`loc` and `loc_container` must both be strings or both be Locators"
            )

        super().__init__(
            page,
            id=id,
            loc=loc,
        )
        self.loc_container = loc_container


class UiWithLabel(UiWithContainer):
    """A mixin class representing UI components with a label."""

    loc_label: Locator
    """
    Playwright `Locator` for the label of the UI element.
    """

    def __init__(
        self,
        page: Page,
        *,
        id: str,
        loc: InitLocator,
        loc_container: InitLocator = "div.shiny-input-container",
        loc_label: InitLocator | None = None,
    ) -> None:
        """
        Initializes the input with a label.

        Parameters
        ----------
        page
            The page where the input is located.
        id
            The id of the UI element.
        loc
            Playwright `Locator` of the UI element.
        loc_container
            Playwright `Locator` of the container of the UI element.
        loc_label
            Playwright `Locator` of the label of the UI element. Defaults to `None`.
        """
        super().__init__(
            page,
            id=id,
            loc_container=loc_container,
            loc=loc,
        )

        if loc_label is None:
            loc_label = f"label#{id}-label"
        if isinstance(loc_label, str):
            loc_label = self.loc_container.locator(loc_label)
        self.loc_label = loc_label

    def expect_label(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the label of the input to have a specific text.

        Parameters
        ----------
        value
            The expected text value of the label.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        playwright_expect(self.loc_label).to_have_text(value, timeout=timeout)


class WidthLocStlyeM:
    """
    A mixin class that provides methods to control the width of input action buttons and action links.

    """

    def expect_width(
        self: UiBaseP,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `width` attribute of a DOM element to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `width` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc, "width", value, timeout=timeout)


class WidthLocM:
    """
    A mixin class representing the `.loc`'s width.

    This class provides methods to expect the width attribute of a DOM element.
    """

    def expect_width(
        self: UiBaseP,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `width` attribute of a DOM element to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `width` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "width", value=value, timeout=timeout)


# # Currently not being used, hence commenting
# class WidthContainerM:
#     """
#     A mixin class representing the container's width.

#     This class provides methods to expect the width attribute of a DOM element's container.
#     """

#     def expect_width(
#         self: UiWithContainerP,
#         value: AttrValue,
#         *,
#         timeout: Timeout = None,
#     ) -> None:
#         """
#         Expect the `width` attribute of a input's container to have a specific value.

#         Parameters
#         ----------
#         value
#             The expected value of the `width` attribute.
#         timeout
#             The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
#         """
#         _expect_attribute_to_have_value(
#             self.loc_container, "width", value=value, timeout=timeout
#         )


class WidthContainerStyleM:
    """
    A mixin class that provides methods to control the width of input elements, such as checkboxes, sliders and radio buttons.
    """

    def expect_width(
        self: UiWithContainerP,
        value: AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the input element to have a specific width.

        Parameters
        ----------
        value
            The expected width.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc_container, "width", value, timeout=timeout)


class InputActionBase(UiBase):
    def expect_label(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the label of the input button to have a specific value.

        Note: This must include the icon if it is present!

        Parameters
        ----------
        value
            The expected value of the label.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """

        self.expect.to_have_text(value, timeout=timeout)

    def click(self, *, timeout: Timeout = None, **kwargs: object) -> None:
        """
        Clicks the input action.

        Parameters
        ----------
        timeout
            The maximum time to wait for the input action to be clicked. Defaults to `None`.
        """
        self.loc.click(timeout=timeout, **kwargs)  # pyright: ignore[reportArgumentType]


Resize = Literal["none", "both", "horizontal", "vertical"]


# * click:
#     * input_checkbox_group
#     * input_radio_buttons


######################################################
# # Outputs
######################################################


class OutputBaseP(Protocol):
    id: str
    loc: Locator
    page: Page



import random

from shiny import reactive
from shiny.express import input, ui

choices = ["A", "B", "C", "D", "E"]
random.shuffle(choices)

ui.input_action_button(
    "remove_panel",
    f"Remove Section {choices[-1]}",
    class_="mt-3 mb-3",
)

" (Sections randomly picked at server start)"

with ui.accordion(id="acc", multiple=True):
    for letter in "ABCDE":
        with ui.accordion_panel(f"Section {letter}"):
            f"Some narrative for section {letter}"


user_choices = [choice for choice in choices]


@reactive.effect
@reactive.event(input.remove_panel)
def _():
    if len(user_choices) == 0:
        ui.notification_show("No more panels to remove!")
        return

    ui.remove_accordion_panel("acc", f"Section {user_choices.pop()}")

    label = "No more panels to remove!"
    if len(user_choices) > 0:
        label = f"Remove Section {user_choices[-1]}"
    ui.update_action_button("remove_panel", label=label)



from __future__ import annotations

import pathlib
from typing import Literal, Optional

from playwright.sync_api import FilePayload, Locator, Page
from playwright.sync_api import expect as playwright_expect

# Import `shiny`'s typing extentions.
# Since this is a private file, tell pyright to ignore the import
from .._types import AttrValue, PatternOrStr, StyleValue, Timeout
from ..expect._internal import (
    expect_attribute_to_have_value as _expect_attribute_to_have_value,
)
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value
from ._base import (
    InputActionBase,
    UiBase,
    UiWithLabel,
    WidthLocStlyeM,
    _expect_multiple,
)


class InputActionButton(
    WidthLocStlyeM,
    InputActionBase,
):
    """Controller for :func:`shiny.ui.input_action_button`."""

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initializes the input action button.

        Parameters
        ----------
        page
            The page where the input action button is located.
        id
            The id of the input action button.
        """
        super().__init__(
            page,
            id=id,
            loc=f"button#{id}.action-button.shiny-bound-input",
        )

    def expect_disabled(self, value: bool, *, timeout: Timeout = None):
        """
        Expect the input action button to be disabled.

        Parameters
        ----------
        value
            The expected value of the `disabled` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "disabled", "" if value else None, timeout=timeout
        )


class InputDarkMode(UiBase):
    """Controller for :func:`shiny.ui.input_dark_mode`."""

    def __init__(
        self,
        page: Page,
        id: Optional[str] | None,
    ) -> None:
        """
        Initializes the input dark mode.

        Parameters
        ----------
        page
            The page where the input dark mode is located.
        id
            The id of the input dark mode.
        """
        id_selector = "" if id is None else f"#{id}"

        super().__init__(
            page,
            id="" if id is None else id,
            loc=f"bslib-input-dark-mode{id_selector}",
        )

    def click(self, *, timeout: Timeout = None):
        """
        Clicks the input dark mode.

        Parameters
        ----------
        timeout
            The maximum time to wait for the input dark mode to be clicked. Defaults to `None`.
        """
        self.loc.click(timeout=timeout)
        return self

    def expect_mode(self, value: str, *, timeout: Timeout = None):
        """
        Expect the `mode` attribute of the input dark mode to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `mode` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "mode", value=value, timeout=timeout)
        self.expect_page_mode(value, timeout=timeout)
        return self

    def expect_page_mode(self, value: str, *, timeout: Timeout = None):
        """
        Expect the page to have a specific dark mode value.

        Parameters
        ----------
        value
            The expected value of the page's dark mode.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.page.locator("html"), "data-bs-theme", value=value, timeout=timeout
        )
        return self

    def expect_attribute(self, value: str, *, timeout: Timeout = None):
        """
        Expect the attribute named `attribute` of the input dark mode to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `attribute` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc, "attribute", value=value, timeout=timeout
        )
        return self


class InputTaskButton(
    WidthLocStlyeM,
    InputActionBase,
):
    """Controller for :func:`shiny.ui.input_task_button`."""

    # TODO-Karan: Test auto_reset functionality
    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initializes the input task button.

        Parameters
        ----------
        page
            The page where the input task button is located.
        id
            The id of the input task button.
        """
        super().__init__(
            page,
            id=id,
            loc=f"button#{id}.bslib-task-button.shiny-bound-input",
        )

    def expect_state(
        self,
        value: Literal["ready", "busy"] | str,
        *,
        timeout: Timeout = None,
    ):
        """
        Expect the state of the input task button to have a specific value.

        Parameters
        ----------
        value
            The expected value of the state of the input task button.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc.locator("> bslib-switch-inline"),
            name="case",
            value=value,
            timeout=timeout,
        )

    def expect_label(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expect the label of the input task button to have a specific value.

        Parameters
        ----------
        value
            The expected value of the label.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        self.expect_label_ready(value, timeout=timeout)

    def expect_label_ready(self, value: PatternOrStr, *, timeout: Timeout = None):
        """
        Expect the label of a ready input task button to have a specific value.

        Parameters
        ----------
        value
            The expected value of the label.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        self.expect_label_state("ready", value, timeout=timeout)

    def expect_label_busy(self, value: PatternOrStr, *, timeout: Timeout = None):
        """
        Expect the label of a busy input task button to have a specific value.

        Parameters
        ----------
        value
            The expected value of the label.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        self.expect_label_state("busy", value, timeout=timeout)

    def expect_label_state(
        self,
        state: str,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ):
        """
        Expect the label of the input task button to have a specific value in a specific state.

        Parameters
        ----------
        state
            The state of the input task button.
        value
            The expected value of the label.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        playwright_expect(
            self.loc.locator(f"> bslib-switch-inline > span[slot='{state}']")
        ).to_have_text(value, timeout=timeout)

    def expect_auto_reset(self, value: bool, timeout: Timeout = None):
        """
        Expect the `auto-reset` attribute of the input task button to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `auto-reset` attribute.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_attribute_to_have_value(
            self.loc,
            name="data-auto-reset",
            value="" if value else None,
            timeout=timeout,
        )


class InputActionLink(InputActionBase):
    """Controller for :func:`shiny.ui.input_action_link`."""

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initializes the input action link.

        Parameters
        ----------
        page
            The page where the input action link is located.
        id
            The id of the input action link.
        """
        super().__init__(
            page,
            id=id,
            loc=f"a#{id}.action-button.shiny-bound-input",
        )


class InputFile(
    # _ExpectPlaceholderAttrM,
    UiWithLabel,
):
    """Controller for :func:`shiny.ui.input_file`."""

    loc_button: Locator
    """
    Playwright `Locator` of the button.
    """
    loc_file_display: Locator
    """
    Playwright `Locator` of the file display.
    """
    loc_progress: Locator
    """
    Playwright `Locator` of the progress bar.
    """

    def __init__(
        self,
        page: Page,
        id: str,
    ) -> None:
        """
        Initialize the InputFile.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The id of the file input.
        """
        super().__init__(
            page,
            id=id,
            loc=f"input[type=file]#{id}",
        )
        self.loc_button = self.loc_container.locator("label span.btn")
        self.loc_file_display = self.loc_container.locator("input[type=text]")
        self.loc_progress = self.loc_container.locator(".progress-bar")

    def set(
        self,
        file_path: (
            str
            | pathlib.Path
            | FilePayload
            | list[str | pathlib.Path]
            | list[FilePayload]
        ),
        *,
        timeout: Timeout = None,
        expect_complete_timeout: Timeout = 30 * 1000,
    ) -> None:
        """
        Set the file upload.

        Parameters
        ----------
        file_path
            The path to the file to upload.
        timeout
            The timeout for the action. Defaults to `None`.
        expect_complete_timeout
            The timeout for the expectation that the upload is complete. Defaults to `30 * 1000`.
        """
        self.loc.wait_for(state="visible", timeout=timeout)
        self.loc.scroll_into_view_if_needed(timeout=timeout)
        self.loc.set_input_files(file_path, timeout=timeout)
        if expect_complete_timeout is not None:
            self.expect_complete(timeout=expect_complete_timeout)

    # TODO-future: Let's make sure that if the upload errors out, expect_complete() fails.
    def expect_complete(
        self,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the file upload to be complete.

        Parameters
        ----------
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc_progress, "width", "100%", timeout=timeout)

    # TODO-future; Test multiple file upload
    def expect_multiple(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expect the `multiple` attribute to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `multiple` attribute.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_multiple(self.loc, value, timeout=timeout)

    def expect_accept(
        self,
        value: list[str] | AttrValue,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `accept` attribute to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `accept` attribute.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        if isinstance(value, list):
            value = ",".join(value)
        _expect_attribute_to_have_value(self.loc, "accept", value, timeout=timeout)

    def expect_width(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expect the width of the input file to have a specific value.

        Parameters
        ----------
        value
            The expected value of the width.
        timeout
            The maximum time to wait for the expectation to be fulfilled. Defaults to `None`.
        """
        _expect_style_to_have_value(self.loc_container, "width", value, timeout=timeout)

    def expect_button_label(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the button label to have a specific value.

        Parameters
        ----------
        value
            The expected value of the button label.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        playwright_expect(self.loc_button).to_have_text(value, timeout=timeout)

    def expect_capture(
        self,
        value: Literal["environment", "user"] | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expect the `capture` attribute to have a specific value.

        Parameters
        ----------
        value
            The expected value of the `capture` attribute.
        timeout
            The timeout for the expectation. Defaults to `None`.
        """
        _expect_attribute_to_have_value(self.loc, "capture", value, timeout=timeout)

    def expect_placeholder(self, value: AttrValue, *, timeout: Timeout = None) -> None:
        _expect_attribute_to_have_value(
            self.loc_file_display, "placeholder", value=value, timeout=timeout
        )



import seaborn as sns
from faicons import icon_svg

# Import data from shared.py
from shared import app_dir, df

from shiny import App, reactive, render, ui

app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.input_slider("mass", "Mass", 2000, 6000, 6000),
        ui.input_checkbox_group(
            "species",
            "Species",
            ["Adelie", "Gentoo", "Chinstrap"],
            selected=["Adelie", "Gentoo", "Chinstrap"],
        ),
        title="Filter controls",
    ),
    ui.layout_column_wrap(
        ui.value_box(
            "Number of penguins",
            ui.output_text("count"),
            showcase=icon_svg("earlybirds"),
        ),
        ui.value_box(
            "Average bill length",
            ui.output_text("bill_length"),
            showcase=icon_svg("ruler-horizontal"),
        ),
        ui.value_box(
            "Average bill depth",
            ui.output_text("bill_depth"),
            showcase=icon_svg("ruler-vertical"),
        ),
        fill=False,
    ),
    ui.layout_columns(
        ui.card(
            ui.card_header("Bill length and depth"),
            ui.output_plot("length_depth"),
            full_screen=True,
        ),
        ui.card(
            ui.card_header("Penguin data"),
            ui.output_data_frame("summary_statistics"),
            full_screen=True,
        ),
    ),
    ui.include_css(app_dir / "styles.css"),
    title="Penguins dashboard",
    fillable=True,
)


def server(input, output, session):
    @reactive.calc
    def filtered_df():
        filt_df = df[df["species"].isin(input.species())]
        filt_df = filt_df.loc[filt_df["body_mass_g"] < input.mass()]
        return filt_df

    @render.text
    def count():
        return filtered_df().shape[0]

    @render.text
    def bill_length():
        return f"{filtered_df()['bill_length_mm'].mean():.1f} mm"

    @render.text
    def bill_depth():
        return f"{filtered_df()['bill_depth_mm'].mean():.1f} mm"

    @render.plot
    def length_depth():
        return sns.scatterplot(
            data=filtered_df(),
            x="bill_length_mm",
            y="bill_depth_mm",
            hue="species",
        )

    @render.data_frame
    def summary_statistics():
        cols = [
            "species",
            "island",
            "bill_length_mm",
            "bill_depth_mm",
            "body_mass_g",
        ]
        return render.DataGrid(filtered_df()[cols], filters=True)


app = App(app_ui, server)



from __future__ import annotations

from typing import (
    AbstractSet,
    Any,
    Collection,
    Literal,
    Protocol,
    Union,
    runtime_checkable,
)


# A duck type for tiktoken.Encoding
class TiktokenEncoding(Protocol):
    name: str

    def encode(
        self,
        text: str,
        *,
        allowed_special: Union[Literal["all"], AbstractSet[str]] = set(),  # noqa: B006
        disallowed_special: Union[Literal["all"], Collection[str]] = "all",
    ) -> list[int]: ...


# A duck type for tokenizers.Encoding
@runtime_checkable
class TokenizersEncoding(Protocol):
    @property
    def ids(self) -> list[int]: ...


# A duck type for tokenizers.Tokenizer
class TokenizersTokenizer(Protocol):
    def encode(
        self,
        sequence: Any,
        pair: Any = None,
        is_pretokenized: bool = False,
        add_special_tokens: bool = True,
    ) -> TokenizersEncoding: ...


TokenEncoding = Union[TiktokenEncoding, TokenizersTokenizer]


def get_default_tokenizer() -> TokenizersTokenizer | None:
    try:
        from tokenizers import Tokenizer

        return Tokenizer.from_pretrained("bert-base-cased")  # type: ignore
    except Exception:
        pass

    return None



import random

from shiny import App, Inputs, Outputs, Session, reactive, ui


def make_panel(letter: str) -> ui.AccordionPanel:
    return ui.accordion_panel(
        f"Section {letter}", f"Some narrative for section {letter}"
    )


items = [make_panel(letter) for letter in "ABCDE"]

app_ui = ui.page_fluid(
    ui.input_action_button("add_panel", "Add random panel", class_="mt-3 mb-3"),
    ui.accordion(*items, id="acc", multiple=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.add_panel)
    def _():
        ui.insert_accordion_panel("acc", make_panel(str(random.randint(0, 10000))))


app = App(app_ui, server)



from __future__ import annotations

__all__ = ("input_handlers",)

from datetime import date, datetime, timezone
from typing import TYPE_CHECKING, Any, Callable, Dict

if TYPE_CHECKING:
    from .session import Session

from .module import ResolvedId
from .types import ActionButtonValue

InputHandlerType = Callable[[Any, ResolvedId, "Session"], Any]


class _InputHandlers(Dict[str, InputHandlerType]):
    def __init__(self):
        super().__init__()

    def add(self, type: str, force: bool = False) -> Callable[[InputHandlerType], None]:
        def _(func: InputHandlerType):
            if type in self and not force:
                raise ValueError(f"Input handler {type} already registered")
            self[type] = func
            return None

        return _

    def remove(self, type: str) -> None:
        del self[type]

    def _process_value(
        self, type: str, value: Any, name: ResolvedId, session: Session
    ) -> Any:
        handler = self.get(type)
        if handler is None:
            raise ValueError("No input handler registered for type: " + type)
        return handler(value, name, session)


input_handlers: _InputHandlers = _InputHandlers()
input_handlers.__doc__ = """
Manage Shiny input handlers.

Add and/or remove input handlers of a given ``type``. Shiny uses these handlers to
pre-process input values from the client (after being deserialized) before passing them
to the ``input`` argument of an :class:`~shiny.App`'s ``server`` function.

The ``type`` is based on the ``getType()`` JavaScript method on the relevant Shiny
input binding. See `this article <https://shiny.posit.co/articles/js-custom-input.html>`_
for more information on how to create custom input bindings. (The article is about
Shiny for R, but the JavaScript and general principles are the same.)

Methods
--------
add(type: str, force: bool = False) -> Callable[[InputHandlerType], None]
    Register an input handler. This method returns a decorator that registers the
    decorated function as the handler for the given ``type``. This handler should
    accept three arguments:
    - the input ``value``
    - the input ``name``
    - the :class:`~shiny.Session` object
remove(type: str)
    Unregister an input handler.

Note
----
``add()`` ing an input handler will make it persist for the duration of the Python
process (unless Shiny is explicitly reloaded). For that reason, verbose naming is
encouraged to minimize the risk of colliding with other Shiny input binding(s) which
happen to use the same ``type`` (if the binding is bundled with a package, we
recommend the format of "packageName.widgetName").

Example
-------
```{python}
#| eval: false
from shiny.input_handler import input_handlers
@input_handlers.add("mypackage.intify")
def _(value, name, session):
    return int(value)
```

On the Javascript side, the associated input binding must have a corresponding
``getType`` method:

```{python}
#| eval: false
getType: function(el) {
    return "mypackage.intify";
}
```
"""


@input_handlers.add("shiny.date")
def _(
    value: str | list[str] | None, name: ResolvedId, session: Session
) -> date | None | tuple[date | None, date | None]:

    if isinstance(value, str) or value is None:
        return _safe_strptime_date(value)
    else:
        return (
            _safe_strptime_date(value[0]),
            _safe_strptime_date(value[1]),
        )


def _safe_strptime_date(value: str | None) -> date | None:
    if value is None:
        return None
    try:
        return datetime.strptime(value, "%Y-%m-%d").date()
    except ValueError:
        return None


@input_handlers.add("shiny.datetime")
def _(
    value: int | float | list[int] | list[float],
    name: ResolvedId,
    session: Session,
) -> datetime | tuple[datetime, datetime]:
    def as_utc_date(x: int | float) -> datetime:
        dt = datetime.fromtimestamp(x, timezone.utc)
        # Remove hour offset from print method by removing the timezone
        # Ex: 2021-08-01T00:00:00+00:00 -> 2021-08-01T00:00:00
        # This is done as all dates are in UTC
        return dt.replace(tzinfo=None)

    if isinstance(value, (int, float)):
        return as_utc_date(value)
    return (as_utc_date(value[0]), as_utc_date(value[1]))


@input_handlers.add("shiny.action")
def _(value: int, name: ResolvedId, session: Session) -> ActionButtonValue:
    # TODO: ActionButtonValue() class can probably be removed
    return ActionButtonValue(value)


# The inputs handlers below currently do nothing, but still need to be defined,
# otherwise there will be an error when the input value is handled.


@input_handlers.add("shiny.number")
def _(value: str, name: ResolvedId, session: Session) -> str:
    return value


# TODO: implement when we have bookmarking
@input_handlers.add("shiny.password")
def _(value: str, name: ResolvedId, session: Session) -> str:
    return value


# TODO: implement when we have bookmarking
@input_handlers.add("shiny.file")
def _(value: Any, name: ResolvedId, session: Session) -> Any:
    return value



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_sidebar_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    left_sidebar = controller.Sidebar(page, "sidebar_left")
    output_txt_left = controller.OutputTextVerbatim(page, "state_left")
    left_sidebar.set(True)
    left_sidebar.expect_padding("10px")
    left_sidebar.expect_padding(["10px"])
    left_sidebar.expect_title("Left sidebar")
    left_sidebar.expect_gap("20px")
    left_sidebar.expect_class("text-white", has_class=True)
    left_sidebar.expect_bg_color("dodgerBlue")
    left_sidebar.expect_desktop_state("open")
    left_sidebar.expect_mobile_state("closed")
    left_sidebar.expect_width("200px")
    output_txt_left.expect_value("input.sidebar_left(): True")
    left_sidebar.expect_open(True)
    left_sidebar.set(False)
    output_txt_left.expect_value("input.sidebar_left(): False")
    left_sidebar.expect_handle(True)
    left_sidebar.expect_open(False)
    left_sidebar.loc_handle.click()
    left_sidebar.expect_open(True)
    output_txt_left.expect_value("input.sidebar_left(): True")

    right_sidebar = controller.Sidebar(page, "sidebar_right")
    right_sidebar.expect_padding(["10px", "20px"])
    right_sidebar.expect_bg_color("SlateBlue")
    right_sidebar.expect_mobile_state("open")
    right_sidebar.expect_desktop_state("closed")

    closed_sidebar = controller.Sidebar(page, "sidebar_closed")
    closed_sidebar.expect_padding(["10px", "20px", "30px"])
    closed_sidebar.expect_bg_color("LightCoral")
    closed_sidebar.expect_mobile_state("closed")
    closed_sidebar.expect_desktop_state("closed")

    always_sidebar = controller.Sidebar(page, "sidebar_always")
    always_sidebar.expect_padding(["10px", "20px", "30px", "40px"])
    always_sidebar.expect_bg_color("PeachPuff")
    always_sidebar.expect_open(True)
    always_sidebar.expect_desktop_state("always")
    always_sidebar.expect_mobile_state("always")
    always_sidebar.expect_mobile_max_height("175px")



from shiny import reactive
from shiny.express import input, ui

with ui.sidebar():
    ui.input_slider("controller", "Controller", min=1, max=3, value=1)

with ui.navset_card_tab(id="inTabset"):
    with ui.nav_panel("Panel 1", value="panel1"):
        "Panel 1 content"
    with ui.nav_panel("Panel 2", value="panel2"):
        "Panel 2 content"
    with ui.nav_panel("Panel 3", value="panel3"):
        "Panel 3 content"


@reactive.effect
def _():
    ui.update_navs("inTabset", selected="panel" + str(input.controller()))



from __future__ import annotations

import typing

from shiny import ui
from shiny.express import ui as xui


def assert_annotations(
    core_fn: typing.Callable[..., typing.Any],
    express_fn: typing.Callable[..., typing.Any],
) -> None:
    ui_a = core_fn.__annotations__
    layout_a = express_fn.__annotations__
    keys: list[str] = []
    for key, _ in ui_a.items():
        keys.append(key)
    for key, _ in layout_a.items():
        if key not in keys:
            keys.append(key)
    for key in keys:
        if key == "args":
            assert key in ui_a
            assert key not in layout_a
        elif key == "return":
            ui_val = ui_a[key]
            layout_val = (
                layout_a[key].replace("RecallContextManager[", "").replace("]", "")
            )
            assert layout_val.endswith(ui_val)
        else:
            assert (
                ui_a[key] == layout_a[key]
            ), f"Type annotations for {key} in {core_fn} (Core) don't match {express_fn} (Express)"


def test_sidebar_annotations() -> None:
    assert_annotations(ui.sidebar, xui.sidebar)



from shiny.express import ui

ui.page_opts(
    id="page_fixed_bottom_inverse_bg",
    position="fixed-bottom",
    bg="dodgerBlue",
    inverse=True,
)

with ui.nav_panel("Data"):
    "This page could be used to pick a dataset."

with ui.nav_panel("View"):
    "This page could be used to view the dataset."



import sys

import pytest
from example_apps import get_apps, reruns, reruns_delay, validate_example
from playwright.sync_api import Page

is_windows = sys.platform.startswith("win")


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
@pytest.mark.parametrize("ex_app_path", get_apps("examples"))
def test_examples(page: Page, ex_app_path: str) -> None:

    skip_on_windows_with_timezonefinder(ex_app_path)
    skip_airmass_on_3_9(ex_app_path)

    validate_example(page, ex_app_path)


def skip_on_windows_with_timezonefinder(ex_app_path: str) -> None:
    if not is_windows:
        return
    if ex_app_path != "examples/airmass/app.py":
        return

    try:
        import timezonefinder  # noqa: F401 # pyright: ignore

        # Future proofing: if timezonefinder is actually available on windows, raise an error
        raise RuntimeError(
            "This code believes timezonefinder is not available on windows. Please remove this check if it is no longer true."
        )
    except ImportError:
        pytest.skip(
            "timezonefinder has difficulty compiling on windows. Skipping example app. posit-dev/py-shiny#1651"
        )


def skip_airmass_on_3_9(ex_app_path: str) -> None:
    print(ex_app_path)
    if ex_app_path != "examples/airmass/app.py":
        return

    import sys

    if sys.version_info[:2] != (3, 9):
        return

    try:
        # Astropy loads `numpy` at run time
        import astropy.coordinates as _  # pyright: ignore # noqa: F401
        import astropy.units as __  # pyright: ignore # noqa: F401

        # Future proofing: if astropy is _actually_ loading, raise an error
        raise RuntimeError(
            "This code believes astropy and numpy have difficulty loading on python 3.9. Please remove this check if it is no longer true."
        )
    except AttributeError as e:
        if "numpy" in str(e) and "product" in str(e):
            pytest.skip(
                "astropy and numpy has difficulty loading on python 3.9. Skipping example app: airmass. posit-dev/py-shiny#1678"
            )
            return

        # Future proofing: if the error is not what we expect, raise it
        raise e



from shared import mtcars

from shiny import App, reactive, render, ui

app_ui = ui.page_fillable(
    ui.layout_columns(
        ui.card(
            ui.card_header(
                ui.markdown(
                    """
                    ##### Editable data frame
                    * Edit the cells!
                    * Sort the columns!
                    """
                )
            ),
            ui.output_data_frame("df_original"),
        ),
        ui.card(
            ui.card_header(
                ui.markdown(
                    """
                    ##### Updated data from the first data frame
                    * Select the rows!
                    * Filter and sort the columns!
                    """
                )
            ),
            ui.output_data_frame("df_edited"),
        ),
        ui.card(
            ui.card_header(
                ui.markdown(
                    """
                    ##### Selected data from the second data frame
                    * Sort the columns!
                    """
                )
            ),
            ui.output_data_frame("df_selected"),
        ),
        col_widths=[4, 4, 4],
    ),
)


def server(input, output, session):
    df = reactive.value(mtcars.iloc[:, range(4)])

    @render.data_frame
    def df_original():
        return render.DataGrid(
            df(),
            editable=True,
        )

    # Convert edited values to the correct data type
    @df_original.set_patch_fn
    def _(*, patch: render.CellPatch) -> render.CellValue:
        if patch["column_index"] in [0, 2]:
            return float(patch["value"])
        return int(patch["value"])

    @render.data_frame
    def df_edited():
        return render.DataGrid(
            # Reactive value is updated when the user edits the data within `df_original` output
            df_original.data_view(),
            selection_mode="rows",
            filters=True,
        )

    @render.data_frame
    def df_selected():
        return render.DataGrid(
            # Reactive value is updated when the user selects rows the data within `df_edited` output
            df_edited.data_view(selected=True),
            selection_mode="rows",
        )


app = App(app_ui, server)



from __future__ import annotations

from playwright.sync_api import Locator
from playwright.sync_api import expect as playwright_expect

from ..._docstring import no_example
from .._types import AttrValue, StyleValue, Timeout
from ._expect import (
    expect_not_to_have_attribute,
    expect_not_to_have_class,
    expect_not_to_have_style,
    expect_to_have_class,
    expect_to_have_style,
)

# NOTE:
# Internal methods only!
# These methods are not meant to be used directly by the user. They are internal methods
# used by the other expect methods to simplify the code and reduce redundancy.


@no_example()
def expect_attribute_to_have_value(
    loc: Locator,
    name: str,
    value: AttrValue,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect an attribute to have a value.

    This method wraps Playwright's Locator expectation `to_have_attribute()` when `value is not `None` and `not_to_have_attribute()` when `value` is `None`. When `value` is `None`, the attribute should not exist and no value should be present.

    Parameters
    ----------
    loc
        The Playwright locator to check.
    name
        The attribute name.
    value
        The attribute value to check. When `value` is `None`, the attribute should not
        exist and no value should be present.
    timeout
        The maximum time to wait for the attribute to appear
    """
    if value is None:
        # Not allowed to have any value for the attribute
        expect_not_to_have_attribute(loc, name, timeout=timeout)
        return

    playwright_expect(loc).to_have_attribute(name=name, value=value, timeout=timeout)


def expect_style_to_have_value(
    loc: Locator,
    key: str,
    # Str representation for value. Will be put in a regex with `key`
    value: StyleValue,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect the `style` attribute to have a value.

    Parameters
    ----------
    loc
        The Playwright locator to check.
    key
        The CSS key to check.
    value
        The CSS value to check. If `None`, then the style attribute should not exist.
    timeout
        The maximum time to wait for the style to appear.
    """
    if value is None:
        expect_not_to_have_style(loc, key, timeout=timeout)
        return

    expect_to_have_style(loc, key, value, timeout=timeout)


def expect_class_to_have_value(
    loc: Locator,
    class_: str,
    *,
    has_class: bool,
    timeout: Timeout = None,
) -> None:
    """Expect a locator to have (or not to have) a class value"""
    if has_class:
        expect_to_have_class(loc, class_, timeout=timeout)
    else:
        expect_not_to_have_class(loc, class_, timeout=timeout)


def _expect_nav_to_have_header_footer(
    parent_loc: Locator,
    header_id: str,
    footer_id: str,
    *,
    timeout: Timeout = None,
) -> None:
    """
    Expect the DOM structure for a header and footer to be preserved.

    Parameters
    ----------
    parent_loc
        The parent locator to check.
    header_id
        The ID of the header element.
    footer_id
        The ID of the footer element.
    timeout
        The maximum time to wait for the header and footer to appear.
    """
    # assert the DOM structure for page_navbar with header and footer is preserved
    class_attr = parent_loc.get_attribute("class")
    if class_attr and "card" in class_attr:
        complicated_parent_loc = parent_loc.locator(
            "xpath=.",
            has=parent_loc.locator("..").locator(
                f".card-body:has(#{header_id}) + .card-body:has(.tab-content) + .card-body #{footer_id}"
            ),
        )
    else:
        complicated_parent_loc = parent_loc.locator(
            "xpath=.",
            has=parent_loc.locator(f"#{header_id} + .tab-content + #{footer_id}"),
        ).locator("..")
    playwright_expect(complicated_parent_loc).to_have_count(1, timeout=timeout)



# pyright: reportUnusedExpression=false
# flake8: noqa
from __future__ import annotations

import contextlib
import inspect
import sys as sys1
from typing import Generator, cast

import pytest
from htmltools import Tagifiable

from shiny import render, ui
from shiny.express import expressify


@contextlib.contextmanager
def capture_display() -> Generator[list[object], None, None]:
    old_displayhook = sys1.displayhook
    displayed: list[object] = []
    sys1.displayhook = displayed.append
    try:
        yield displayed
    finally:
        sys1.displayhook = old_displayhook


@expressify()
def display_repeated(value: str, /, times: int, *, sep: str = " ") -> None:
    sep.join([value] * times)


def test_simple():
    with capture_display() as d:
        display_repeated("hello", 3)
        assert d == ["hello hello hello"]

    with capture_display() as d:
        display_repeated("hello", 3, sep=", ")
        assert d == ["hello, hello, hello"]

    with capture_display() as d:
        display_repeated("hello", times=3, sep=", ")
        assert d == ["hello, hello, hello"]

    with pytest.raises(TypeError):
        display_repeated("hello")  # type: ignore


@expressify()
def display_variadic(*args: object, **kwargs: object):
    "# args"
    for arg in args:
        arg
    "# kwargs"
    for key, value in kwargs.items():
        (key, value)


def test_null_filtered():
    @expressify()
    def has_none():
        1
        None
        2

    with capture_display() as d:
        has_none()
        assert d == [1, 2]


def test_variadic():
    with capture_display() as d:
        display_variadic("one", "two", three="four")
        assert d == ["# args", "one", "two", "# kwargs", ("three", "four")]


def nested(z: int = 1):
    x = 2

    @expressify()
    def inner():
        x * 10 * z

    with capture_display() as d:
        inner()
        assert d == [x * 10 * z]


def test_caching():
    import shiny.express.expressify_decorator._expressify as _expressify

    nested()
    cache_len_before = len(_expressify.code_cache)
    nested(z=3)
    cache_len_after = len(_expressify.code_cache)
    assert cache_len_before == cache_len_after


def test_duplicate_func_names_ok():
    """
    The two inner() functions should be treated as different from each other,
    and different from the third copy in test_nested()
    """

    x = "hello"

    @expressify()
    def inner():  # pyright: ignore[reportRedeclaration]
        x + " world"

    inner_old = inner

    @expressify()
    def inner():  # pyright: ignore[reportRedeclaration]
        x + " universe"

    with capture_display() as d:
        inner_old()
        assert d == ["hello world"]

    with capture_display() as d:
        inner()
        assert d == ["hello universe"]

    # Here's yet another one, just to be mean
    @expressify()
    def inner():
        x + " nobody"


def test_not_decorated():
    def not_decorated():
        1
        2
        3

    decorated = expressify()(not_decorated)

    with capture_display() as d:
        decorated()
        assert d == [1, 2, 3]

    with capture_display() as d:
        not_decorated()
        assert d == []

    assert inspect.getsource(decorated) == inspect.getsource(not_decorated)


def test_annotations():
    @expressify()
    def annotated(x: int, y: int) -> int:
        """Here's a docstring"""
        x + y
        return 0

    assert annotated.__name__ == "annotated"
    assert annotated.__annotations__ == {"x": "int", "y": "int", "return": "int"}
    assert annotated.__doc__ == "Here's a docstring"

    assert inspect.getsource(annotated) == inspect.getsource(annotated.__wrapped__)  # type: ignore


def test_implicit_output():
    @expressify()
    def has_implicit_outputs():
        @render.code
        def foo():
            return "hello"

    with capture_display() as d:
        has_implicit_outputs()
        assert len(d) == 1
        d0 = cast(Tagifiable, d[0])
        assert d0.tagify() == ui.output_code("foo")


def test_no_nested_transform_unless_explicit():
    @expressify()
    def inner1():
        1
        2
        None

        def inner2():
            # Doesn't transform, it doesn't have the decorator
            3
            4

            @expressify()
            def inner3():
                # Does transform, it has the decorator again
                5
                6

            inner3()

        inner2()

    with capture_display() as d:
        inner1()
        assert d == [1, 2, 5, 6]



from __future__ import annotations

from htmltools import Tag, css

from shiny.express import ui
from shiny.express.ui import fill

ui.markdown(
    """\
    # `as_fillable_container()`

    For an item to fill its parent element,
    * the item must have `as_fill_item()` be called on it
    * the parent container must have `as_fillable_container()` called on it

    If both methods are called, the inner child will naturally expand into its parent container.
    """
)


def outer_inner() -> Tag:
    inner = ui.div(
        id="inner",
        style=css(
            height="200px",
            border="3px blue solid",
        ),
    )
    outer = ui.div(
        inner,
        id="outer",
        style=css(
            height="300px",
            border="3px red solid",
        ),
    )
    return outer


outer0 = outer_inner()

outer1 = outer_inner()
outer1.children[0] = fill.as_fill_item(outer1.children[0])

outer2 = outer_inner()
outer2 = fill.as_fillable_container(outer2)
outer2.children[0] = fill.as_fill_item(outer2.children[0])

with ui.layout_columns():
    ui.markdown("##### Default behavior")
    ui.markdown("##### `as_fill_item(blue)`")
    ui.markdown("##### `as_fill_item(blue)` + `as_fillable_container(red)`")

with ui.layout_columns():
    outer0
    outer1
    outer2



# ------------------------------------------------------------------------------------
# When putting a Chat into production, there are at least a couple additional
# considerations to keep in mind:
#  - Token Limits: LLMs have (varying) limits on how many tokens can be included in
#    a single request and response. To accurately respect these limits, you'll want
#    to find the revelant limits and tokenizer for the model you're using, and inform
#    Chat about them.
#  - Reproducibility: Consider pinning a snapshot of the LLM model to ensure that the
#    same model is used each time the app is run.
#
# See the MODEL_INFO dictionary below for an example of how to set these values for
# OpenAI's GPT-4o model.
# ------------------------------------------------------------------------------------
import os

import tiktoken
from app_utils import load_dotenv
from openai import AsyncOpenAI

from shiny.express import ui

load_dotenv()
llm = AsyncOpenAI(api_key=os.environ.get("OPENAI_API_KEY"))


MODEL_INFO = {
    "name": "gpt-4o-2024-08-06",
    "tokenizer": tiktoken.encoding_for_model("gpt-4o-2024-08-06"),
    "token_limits": (128000, 16000),
}


ui.page_opts(
    title="Hello OpenAI Chat",
    fillable=True,
    fillable_mobile=True,
)

chat = ui.Chat(
    id="chat",
    messages=[
        {"content": "Hello! How can I help you today?", "role": "assistant"},
    ],
    tokenizer=MODEL_INFO["tokenizer"],
)

chat.ui()


@chat.on_user_submit
async def _():
    messages = chat.messages(format="openai", token_limits=MODEL_INFO["token_limits"])
    response = await llm.chat.completions.create(
        model=MODEL_INFO["name"], messages=messages, stream=True
    )
    await chat.append_message_stream(response)



# pyright: reportUnusedFunction=false

import typing

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.input_numeric("n", "N", 20),
    ui.input_numeric("n2", "N2", 50),
    ui.input_checkbox("checkbox", "Checkbox", True),
    ui.output_code("txt", placeholder=True),
    ui.output_code("txt2", placeholder=True),
    ui.output_code("txt3", placeholder=True),
)


# By default the type of any input value, like input.n(), is Any, so no type checking
# will be used.
#
# But it is possible to specify the type of the input value, by creating a subclass of
# Inputs. We'll do that for input.n2() and input.checkbox():
class ShinyInputs(Inputs):
    n2: reactive.value[int]
    check: reactive.value[bool]


def server(input: Inputs, output: Outputs, session: Session):
    # Cast `input` to our ShinyInputs class. This just tells the static type checker
    # that we want it treated as a ShinyInputs object for type checking; it has no
    # run-time effect.
    input = typing.cast(ShinyInputs, input)

    # The type checker knows that r() returns an int, which you can see if you hover
    # over it.
    @reactive.calc
    def r():
        if input.n() is None:
            return 0
        return input.n() * 2

    # Because we did NOT add a type for input.n (we did input.n2), the type checker
    # thinks the return type of input.n() is Any, so we don't get type checking here.
    # The function is returning the wrong value here: it returns an int instead of a
    # string, but this error is not flagged.
    @render.code
    async def txt():
        return input.n() * 2

    # In contrast, input.n2() is declared to return an int, so the type check does flag
    # this error -- the `render.code()` is underlined in red.
    @render.code
    async def txt2():
        return input.n2() * 2

    # This is a corrected version of the function above. It returns a string, and is not
    # marked in red.
    @render.code
    async def txt3():
        return str(input.n2() * 2)


app = App(app_ui, server)



from datetime import date

from shiny import App, Inputs, Outputs, Session, ui

app_ui = ui.page_fluid(
    ui.input_date("date1", "Has default date:", value="2016-02-29"),
    # Default value is the date in client's time zone
    ui.input_date("date2", "Client's current date:"),
    # value is always yyyy-mm-dd, even if the display format is different
    ui.input_date("date3", "Format mm/dd/yy:", value="2016-02-29", format="mm/dd/yy"),
    # Pass in a Date object
    ui.input_date("date4", "Default uses date object:", value=date(2016, 2, 29)),
    # Use different language and different first day of week
    ui.input_date(
        "date5",
        "Language is German and the week starts on Monday:",
        language="ru",
        weekstart=1,
    ),
    # Start with decade view instead of default month view
    ui.input_date("date6", "Start Date picker in Decade view:", startview="decade"),
    # Disable Mondays and Tuesdays.
    ui.input_date("date7", "Disable Monday and Tuesday:", daysofweekdisabled=[1, 2]),
    # Disable specific dates.
    ui.input_date(
        "date8",
        "Disable specific dates:",
        value="2016-02-29",
        datesdisabled=["2016-03-01", "2016-03-02"],
    ),
    # Set min and max dates.
    ui.input_date(
        "date9",
        "Set min and max dates:",
        value="2016-02-03",
        min="2016-02-01",
        max="2016-02-29",
    ),
    # Set width of the date field
    ui.input_date("date10", "Set width of text input:", width="600px"),
    # Set autoclose to false
    ui.input_date("date11", "Auto close is disabled:", autoclose=False),
)


def server(input: Inputs, output: Outputs, session: Session):
    pass


app = App(app_ui, server)



import random
import time

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.card(
        ui.layout_columns(
            ui.input_action_button("first", "Invalidate first (slow) computation"),
            ui.input_action_button("second", "Invalidate second (fast) computation"),
        ),
        ui.output_text_verbatim("result"),
    )
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc
    def first():
        input.first()
        p = ui.Progress()
        for i in range(30):
            p.set(i / 30, message="Computing, please wait...")
            time.sleep(0.1)
        p.close()
        return random.randint(1, 1000)

    @reactive.calc
    def second():
        input.second()
        return random.randint(1, 1000)

    @render.text
    def result():
        return first() + second()


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, render, ui


def make_items():
    return [
        ui.accordion_panel(f"Section {letter}", f"Some narrative for section {letter}")
        for letter in "ABCDE"
    ]


# # First shown by default
# ui.accordion(*make_items())

# # Nothing shown by default
# ui.accordion(*make_items(), open=False)
# # Everything shown by default
# ui.accordion(*make_items(), open=True)

# # Show particular sections
# ui.accordion(*make_items(), open="Section B")
# ui.accordion(*make_items(), open=["Section A", "Section B"])


app_ui = ui.page_fluid(
    ui.markdown("#### Accordion: (`multiple=False`)"),
    # Provide an id to create a shiny input binding
    ui.accordion(*make_items(), id="acc_single", multiple=False),
    ui.output_text_verbatim("acc_single_val", placeholder=True),
    ui.tags.br(),
    ui.markdown("#### Accordion: (`multiple=True`)"),
    ui.accordion(*make_items(), id="acc_multiple"),
    ui.output_text_verbatim("acc_multiple_val", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def acc_single_val():
        return "input.acc_single(): " + str(input.acc_single())

    @render.text
    def acc_multiple_val():
        return "input.acc_multiple(): " + str(input.acc_multiple())


app = App(app_ui, server)



import os
import time

import numpy as np
import seaborn as sns

from shiny.express import input, render, ui

ui.page_opts(title="Busy spinner demo")

with ui.sidebar():
    ui.input_selectize(
        "indicator_types",
        "Busy indicator types",
        ["spinners", "pulse"],
        multiple=True,
        selected=["spinners", "pulse"],
    )

    @render.download
    def download():
        time.sleep(3)
        path = os.path.join(os.path.dirname(__file__), "app-express.py")
        return path


with ui.card():
    ui.card_header(
        "Plot that takes a few seconds to render",
        ui.input_task_button("simulate", "Simulate"),
        class_="d-flex justify-content-between align-items-center",
    )

    @render.plot
    def plot():
        input.simulate()
        time.sleep(3)
        sns.lineplot(x=np.arange(100), y=np.random.randn(100))


ui.busy_indicators.options(spinner_type="bars3")


@render.ui
def indicator_types_ui():
    return ui.busy_indicators.use(
        spinners="spinners" in input.indicator_types(),
        pulse="pulse" in input.indicator_types(),
    )



import matplotlib.pyplot as plt
import numpy as np

from shiny import reactive
from shiny.express import input, render, ui

ui.input_slider("n", "Number of observations", min=0, max=1000, value=500)
ui.input_action_button("go", "Go!", class_="btn-success")


@render.plot(alt="A histogram")
def plot():
    # Take a reactive dependency on the action button...
    input.go()

    # ...but don't take a reactive dependency on the slider
    with reactive.isolate():
        np.random.seed(19680801)
        x = 100 + 15 * np.random.randn(input.n())

    fig, ax = plt.subplots()
    ax.hist(x, bins=30, density=True)
    return fig



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_default_page_navbar(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default_page_navbar = controller.PageNavbar(page, "default_page_navbar")
    default_page_navbar.expect_title("Default Page Navbar")
    default_page_navbar.expect_inverse(value=False)
    default_page_navbar.expect_lang("en")
    default_page_navbar.expect_window_title("Page NavBar title")
    default_page_navbar.expect_fillable_mobile(True)



from starlette.requests import Request
from starlette.responses import JSONResponse

from shiny import reactive
from shiny.express import input, session, ui

ui.input_action_button("serve", "Click to serve")

ui.div(id="messages")


@reactive.effect
@reactive.event(input.serve)
def _():
    async def my_handler(request: Request) -> JSONResponse:
        return JSONResponse({"n_clicks": input.serve()}, status_code=200)

    path = session.dynamic_route("my_handler", my_handler)

    print("Serving at: ", path)

    ui.insert_ui(
        ui.tags.script(
            f"""
            fetch('{path}')
                .then(r => r.json())
                .then(x => {{ $('#messages').text(`Clicked ${{x.n_clicks}} times`); }});
            """
        ),
        selector="body",
    )



from shared import mtcars

from shiny import reactive
from shiny.express import render, ui

df = reactive.value(mtcars.iloc[:, range(4)])


with ui.layout_columns(col_widths=[4, 4, 4]):
    with ui.card():
        with ui.card_header():
            ui.markdown(
                """
                ##### Editable data frame
                * Edit the cells!
                * Sort the columns!
                """
            )

            @render.data_frame
            def df_original():
                return render.DataGrid(
                    df(),
                    editable=True,
                )

            # Convert edited values to the correct data type
            @df_original.set_patch_fn
            def _(*, patch: render.CellPatch) -> render.CellValue:
                if patch["column_index"] in [0, 2]:
                    return float(patch["value"])
                return int(patch["value"])

    with ui.card():
        with ui.card_header():
            ui.markdown(
                """
                ##### Updated data from the first data frame
                * Select the rows!
                * Filter and sort the columns!
                """
            )

            @render.data_frame
            def df_edited():
                return render.DataGrid(
                    df_original.data_view(),
                    selection_mode="rows",
                    filters=True,
                )

    with ui.card():
        with ui.card_header():
            ui.markdown(
                """
                ##### Selected data from the second data frame
                * Sort the columns!
                """
            )

            @render.data_frame
            def df_selected():
                return render.DataGrid(
                    df_edited.data_view(selected=True),
                    selection_mode="none",
                )



from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_express_page_fluid(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    txt = controller.OutputTextVerbatim(page, "visible")
    txt.expect_value("40")

    expect(page.locator("#visible")).to_have_count(1)
    expect(page.locator("#hidden")).to_have_count(0)



import matplotlib.pyplot as plt
import numpy as np

from shiny import render
from shiny.express import input, ui

with ui.layout_column_wrap(width=1 / 2):
    with ui.navset_underline():
        with ui.nav_panel(title="One"):
            ui.input_slider("n", "N", 1, 100, 50)

        with ui.nav_panel(title="Two"):

            @render.plot
            def histogram():
                np.random.seed(19680801)
                x = 100 + 15 * np.random.randn(437)
                plt.hist(x, input.n(), density=True)

    with ui.navset_card_underline():
        with ui.nav_panel(title="One"):
            ui.input_slider("n2", "N", 1, 100, 50)

        with ui.nav_panel(title="Two"):

            @render.plot
            def histogram2():
                np.random.seed(19680801)
                x = 100 + 15 * np.random.randn(437)
                plt.hist(x, input.n2(), density=True)



import pandas as pd

from shiny import render
from shiny.express import ui

data = {
    "A": [1, 2, 3, 4, 5, 6],
    "B": ["a", "b", "c", "d", "e", "f"],
    "C": [10.1, 20.2, 30.3, 40.4, 50.5, 60.6],
    "D": ["apple", "banana", "cherry", "date", "elderberry", "fig"],
    "E": [True, False, True, False, True, False],
    "F": ["John", "Jane", "Jim", "Jessie", "Jack", "Jill"],
}

df = pd.DataFrame(data)

ui.page_opts(fillable=True)

with ui.card(id="card"):
    ui.h2("Below is a sample dataframe")

    @render.data_frame
    def sample_data_frame():
        return df



from __future__ import annotations

import datetime

from playwright.sync_api import Page

from shiny.playwright import controller


def expect_state(
    page: Page,
    module_id: str,
    key: str,
    value: object,
):
    new_module_id = f"{module_id}-" if module_id else ""
    controller.OutputTextVerbatim(
        page,
        f"{new_module_id}status_{key}",
    ).expect_value(f"ui.{key}: `{value}`")


def expect_x_state(
    page: Page,
    module_id: str,
    key: str,
    value: object,
):
    new_module_id = f"{module_id}-" if module_id else ""
    new_key = key.replace("x_", "")
    controller.OutputTextVerbatim(
        page,
        f"{new_module_id}status_x_{new_key}",
    ).expect_value(f"x.ui.{new_key}: `{value}`")


def expect_mod_state(
    page: Page,
    module_id: str,
    *,
    sidebar: bool,
    accordion: tuple[str, ...],
    popover: bool,
    tooltip: bool,
    input_action_button: int,
    input_action_link: int,
    input_file: str | None,
    input_checkbox: bool,
    input_checkbox_group: tuple[str, ...],
    input_date: datetime.date | str,
    input_date_range: tuple[datetime.date, datetime.date],
    input_numeric: int,
    input_password: str,
    input_radio_buttons: str,
    input_select: str,
    input_selectize: str,
    input_slider: int,
    input_switch: bool,
    input_text: str,
    input_text_area: str,
    navset_bar: str,
    navset_card_pill: str,
    navset_card_tab: str,
    navset_hidden: str,
    navset_pill: str,
    navset_tab: str,
):
    # print(str(locals().items()))
    # Must realize the locals before iterating
    for key, value in list(locals().items()):
        if key == "page" or key == "module_id":
            continue
        if key.startswith("x_"):
            expect_x_state(page, module_id, key, value)
        else:
            expect_state(page, module_id, key, value)


def expect_default_mod_state(page: Page, module_id: str):
    expect_mod_state(
        page,
        module_id=module_id,
        sidebar=True,
        accordion=("a",),
        popover=False,
        tooltip=False,
        input_action_button=0,
        input_action_link=0,
        input_file=None,
        input_checkbox=False,
        input_checkbox_group=("b", "c", "d"),
        input_date=datetime.date(2023, 8, 24),
        input_date_range=(datetime.date(2023, 8, 25), datetime.date(2023, 8, 27)),
        input_numeric=0,
        input_password="password0",
        input_radio_buttons="a",
        input_select="a",
        input_selectize="a",
        input_slider=0,
        input_switch=False,
        input_text="text0",
        input_text_area="text_area0",
        navset_bar="a",
        navset_card_pill="a",
        navset_card_tab="a",
        navset_hidden="a",
        navset_pill="a",
        navset_tab="a",
    )



from __future__ import annotations

import copy
import importlib
import importlib.util
import inspect
import os
import platform
import re
import sys
import types
from pathlib import Path
from typing import Any, Optional

import click
import uvicorn
import uvicorn.config

import shiny

from . import __version__, _autoreload, _hostenv, _static, _utils
from ._docstring import no_example
from ._typing_extensions import NotRequired, TypedDict
from .express import is_express_app
from .express._utils import escape_to_var_name


@click.group("main")
@click.version_option(__version__)
def main() -> None:
    pass


stop_shortcut = "Ctrl+C"

RELOAD_INCLUDES_DEFAULT = (
    "*.py",
    "*.css",
    "*.scss",
    "*.js",
    "*.htm",
    "*.html",
    "*.png",
    "*.yml",
    "*.yaml",
)
RELOAD_EXCLUDES_DEFAULT = (".*", "*.py[cod]", "__pycache__", "env", "venv")


@main.command(
    help=f"""Run a Shiny app (press {stop_shortcut} to stop).

The APP argument indicates the Python file (or module) and attribute where the
shiny.App object can be found. For example, if the current directory contains
a file called app.py, which contains an `app` variable that is the Shiny app,
any of the following will work:

\b
  * shiny run             # app:app is assumed
  * shiny run app         # :app is assumed
  * shiny run app.py      # :app is assumed
  * shiny run app:app
  * shiny run app.py:app
"""
)
@click.argument("app", default="app.py:app")
@click.option(
    "-h",
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this host.",
    show_default=True,
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=8000,
    help="Bind socket to this port. If 0, a random port will be used.",
    show_default=True,
)
@click.option(
    "--autoreload-port",
    type=int,
    default=0,
    help="Bind autoreload socket to this port. If 0, a random port will be used. Ignored if --reload is not used.",
    show_default=True,
)
@click.option(
    "-r",
    "--reload",
    is_flag=True,
    default=False,
    help="Enable auto-reload. See --reload-includes for types of files that are monitored for changes.",
)
@click.option(
    "--reload-dir",
    "reload_dirs",
    multiple=True,
    help="Indicate a directory `--reload` should (recursively) monitor for changes, in "
    "addition to the app's parent directory. Can be used more than once.",
    type=click.Path(exists=True),
)
@click.option(
    "--reload-includes",
    "reload_includes",
    default=",".join(RELOAD_INCLUDES_DEFAULT),
    help="File glob(s) to indicate which files should be monitored for changes. Defaults"
    f' to "{",".join(RELOAD_INCLUDES_DEFAULT)}".',
)
@click.option(
    "--reload-excludes",
    "reload_excludes",
    default=",".join(RELOAD_EXCLUDES_DEFAULT),
    help="File glob(s) to indicate which files should be excluded from file monitoring. Defaults"
    f' to "{",".join(RELOAD_EXCLUDES_DEFAULT)}".',
)
@click.option(
    "--ws-max-size",
    type=int,
    default=16777216,
    help="WebSocket max size message in bytes",
    show_default=True,
)
@click.option(
    "--log-level",
    type=click.Choice(list(uvicorn.config.LOG_LEVELS.keys())),
    default=None,
    help="Log level. [default: info]",
    show_default=True,
)
@click.option(
    "-d",
    "--app-dir",
    default=".",
    show_default=True,
    help="Look for APP in the specified directory, by adding this to the PYTHONPATH."
    " Defaults to the current working directory. If APP is a file path, this argument"
    " is ignored.",
)
@click.option(
    "--factory",
    is_flag=True,
    default=False,
    help="Treat APP as an application factory, i.e. a () -> <ASGI app> callable.",
    show_default=True,
)
@click.option(
    "-b",
    "--launch-browser",
    is_flag=True,
    default=False,
    help="Launch app browser after app starts, using the Python webbrowser module.",
    show_default=True,
)
@click.option(
    "--dev-mode/--no-dev-mode",
    is_flag=True,
    default=True,
    help="Dev mode",
    show_default=True,
)
@no_example()
def run(
    app: str | shiny.App,
    host: str,
    port: int,
    *,
    autoreload_port: int,
    reload: bool,
    reload_dirs: tuple[str, ...],
    reload_includes: str,
    reload_excludes: str,
    ws_max_size: int,
    log_level: str,
    app_dir: str,
    factory: bool,
    launch_browser: bool,
    dev_mode: bool,
    **kwargs: object,
) -> None:
    reload_includes_list = reload_includes.split(",")
    reload_excludes_list = reload_excludes.split(",")
    return run_app(
        app,
        host=host,
        port=port,
        autoreload_port=autoreload_port,
        reload=reload,
        reload_dirs=list(reload_dirs),
        reload_includes=reload_includes_list,
        reload_excludes=reload_excludes_list,
        ws_max_size=ws_max_size,
        log_level=log_level,
        app_dir=app_dir,
        factory=factory,
        launch_browser=launch_browser,
        dev_mode=dev_mode,
        **kwargs,
    )


def run_app(
    app: str | shiny.App = "app:app",
    host: str = "127.0.0.1",
    port: int = 8000,
    *,
    autoreload_port: int = 0,
    reload: bool = False,
    reload_dirs: Optional[list[str]] = None,
    reload_includes: list[str] | tuple[str, ...] = RELOAD_INCLUDES_DEFAULT,
    reload_excludes: list[str] | tuple[str, ...] = RELOAD_EXCLUDES_DEFAULT,
    ws_max_size: int = 16777216,
    log_level: Optional[str] = None,
    app_dir: Optional[str] = ".",
    factory: bool = False,
    launch_browser: bool = False,
    dev_mode: bool = True,
    **kwargs: object,
) -> None:
    """
    Starts a Shiny app. Press ``Ctrl+C`` (or ``Ctrl+Break`` on Windows) to stop the app.

    Parameters
    ----------
    app
        The app to run. The default, ``app:app``, represents the "usual" case where the
        application is named ``app`` inside a ``app.py`` file within the current working
        directory. In other cases, the app location can be specified as a
        ``<module>:<attribute>`` string where the ``:<attribute>`` is only necessary if
        the application is named something other than ``app``. Note that ``<module>``
        can be a relative path to a ``.py`` file or a directory (with an ``app.py`` file
        inside of it); and in this case, the relative path is resolved relative to the
        ``app_dir`` directory.
    host
        The address that the app should listen on.
    port
        The port that the app should listen on. Set to 0 to use a random port.
    autoreload_port
        The port that should be used for an additional websocket that is used to support
        hot-reload. Set to 0 to use a random port.
    reload
        Enable auto-reload.
    reload_dirs
        A list of directories (in addition to the app directory) to watch for changes that
        will trigger an app reload.
    reload_includes
        List or tuple of file globs to indicate which files should be monitored for
        changes. Can be combined with `reload_excludes`.
    reload_excludes
        List or tuple of file globs to indicate which files should be excluded from
        reload monitoring. Can be combined with `reload_includes`
    ws_max_size
        WebSocket max size message in bytes.
    log_level
        Log level.
    app_dir
        The directory to look for ``app`` under (by adding this to the ``PYTHONPATH``).
    factory
        Treat ``app`` as an application factory, i.e. a () -> <ASGI app> callable.
    launch_browser
        Launch app browser after app starts, using the Python webbrowser module.
    **kwargs
        Additional keyword arguments which are passed to ``uvicorn.run``. For more
        information see [Uvicorn documentation](https://www.uvicorn.org/).

    Tip
    ---
    The ``shiny run`` command-line interface (which comes installed with Shiny) provides
    the same functionality as `shiny.run_app()`.

    Examples
    --------

    ```{python}
    #|eval: false
    from shiny import run_app

    # Run ``app`` inside ``./app.py``
    run_app()

    # Run ``app`` inside ``./myapp.py`` (or ``./myapp/app.py``)
    run_app("myapp")

    # Run ``my_app`` inside ``./myapp.py`` (or ``./myapp/app.py``)
    run_app("myapp:my_app")

    # Run ``my_app`` inside ``../myapp.py`` (or ``../myapp/app.py``)
    run_app("myapp:my_app", app_dir="..")
    ```
    """

    # If port is 0, randomize
    if port == 0:
        port = _utils.random_port(host=host)

    os.environ["SHINY_HOST"] = host
    os.environ["SHINY_PORT"] = str(port)
    if dev_mode:
        os.environ["SHINY_DEV_MODE"] = "1"

    if isinstance(app, str):
        # Remove ":app" suffix if present. Normally users would just pass in the
        # filename without the trailing ":app", as in `shiny run app.py`, but the
        # default value for `shiny run` is "app.py:app", so we need to handle it.
        app_no_suffix = re.sub(r":app$", "", app)
        if is_express_app(app_no_suffix, app_dir):
            app_path = Path(app_no_suffix).resolve()
            # If the file is "/path/to/app.py", our entrypoint with the escaped filename
            # is "shiny.express.app:_2f_path_2f_to_2f_app_2e_py".
            app = "shiny.express.app:" + escape_to_var_name(str(app_path))
            app_dir = str(app_path.parent)

            # Express apps need min version of rsconnect-python to deploy correctly.
            _verify_rsconnect_version()
        else:
            app, app_dir = resolve_app(app, app_dir)

    if app_dir:
        app_dir = os.path.realpath(app_dir)

    log_config: dict[str, Any] = copy.deepcopy(uvicorn.config.LOGGING_CONFIG)

    if reload_dirs is None:
        reload_dirs = []
        if app_dir is not None:
            reload_dirs = [app_dir]

    if reload:
        # Always watch the app_dir
        if app_dir and app_dir not in reload_dirs:
            reload_dirs.append(app_dir)
        # For developers of Shiny itself; autoreload the app when Shiny package changes
        if os.getenv("SHINY_PKG_AUTORELOAD"):
            shinypath = Path(inspect.getfile(shiny)).parent
            reload_dirs.append(str(shinypath))

        if autoreload_port == 0:
            autoreload_port = _utils.random_port(host=host)

        if autoreload_port == port:
            print(
                "Autoreload port is already being used by the app; disabling autoreload\n",
                file=sys.stderr,
            )
            reload = False
        else:
            setup_hot_reload(log_config, autoreload_port, port, launch_browser)

    reload_args: ReloadArgs = {}
    if reload:
        reload_args = {
            "reload": reload,
            # Adding `reload_includes` param while `reload=False` produces an warning
            # https://github.com/encode/uvicorn/blob/d43afed1cfa018a85c83094da8a2dd29f656d676/uvicorn/config.py#L298-L304
            "reload_includes": list(reload_includes),
            "reload_excludes": list(reload_excludes),
            "reload_dirs": reload_dirs,
        }

    if launch_browser and not reload:
        setup_launch_browser(log_config)

    maybe_setup_rsw_proxying(log_config)

    uvicorn.run(  # pyright: ignore[reportUnknownMemberType]
        app,
        host=host,
        port=port,
        ws_max_size=ws_max_size,
        log_level=log_level,
        log_config=log_config,
        app_dir=app_dir,
        factory=factory,
        lifespan="on",
        # Don't allow shiny to use uvloop!
        # https://github.com/posit-dev/py-shiny/issues/1373
        loop="asyncio",
        **reload_args,  # pyright: ignore[reportArgumentType]
        **kwargs,
    )


def setup_hot_reload(
    log_config: dict[str, Any],
    autoreload_port: int,
    app_port: int,
    launch_browser: bool,
) -> None:
    # The only way I've found to get notified when uvicorn decides to reload, is by
    # inserting a custom log handler.
    log_config["handlers"]["shiny_hot_reload"] = {
        "class": "shiny._autoreload.HotReloadHandler",
        "level": "INFO",
    }
    if "handlers" not in log_config["loggers"]["uvicorn.error"]:
        log_config["loggers"]["uvicorn.error"]["handlers"] = []
    log_config["loggers"]["uvicorn.error"]["handlers"].append("shiny_hot_reload")

    _autoreload.start_server(autoreload_port, app_port, launch_browser)


def setup_launch_browser(log_config: dict[str, Any]):
    log_config["handlers"]["shiny_launch_browser"] = {
        "class": "shiny._launchbrowser.LaunchBrowserHandler",
        "level": "INFO",
    }
    if "handlers" not in log_config["loggers"]["uvicorn.error"]:
        log_config["loggers"]["uvicorn.error"]["handlers"] = []
    log_config["loggers"]["uvicorn.error"]["handlers"].append("shiny_launch_browser")


def maybe_setup_rsw_proxying(log_config: dict[str, Any]) -> None:
    # Replace localhost URLs emitted to the log, with proxied URLs
    if _hostenv.is_workbench():
        if "filters" not in log_config:
            log_config["filters"] = {}
        log_config["filters"]["rsw_proxy"] = {"()": "shiny._hostenv.ProxyUrlFilter"}
        if "filters" not in log_config["handlers"]["default"]:
            log_config["handlers"]["default"]["filters"] = []
        log_config["handlers"]["default"]["filters"].append("rsw_proxy")


def is_file(app: str) -> bool:
    return "/" in app or app.endswith(".py")


def resolve_app(app: str, app_dir: str | None) -> tuple[str, str | None]:
    # The `app` parameter can be:
    #
    # - A module:attribute name
    # - An absolute or relative path to a:
    #   - .py file (look for app inside of it)
    #   - directory (look for app:app inside of it)
    # - A module name (look for :app) inside of it

    if platform.system() == "Windows" and re.match("^[a-zA-Z]:[/\\\\]", app):
        # On Windows, need special handling of ':' in some cases, like these:
        #   shiny run c:/Users/username/Documents/myapp/app.py
        #   shiny run c:\Users\username\Documents\myapp\app.py
        module, attr = app, ""
    else:
        module, _, attr = app.partition(":")
    if not module:
        raise ImportError("The APP parameter cannot start with ':'.")
    if not attr:
        attr = "app"

    if is_file(module):
        # Before checking module path, resolve it relative to app_dir if provided
        module_path = module if app_dir is None else os.path.join(app_dir, module)
        # TODO: We should probably be using some kind of loader
        # TODO: I don't like that we exit here, if we ever export this it would be bad;
        #       but also printing a massive stack trace for a `shiny run badpath` is way
        #       unfriendly. We should probably throw a custom error that the shiny run
        #       entrypoint knows not to print the stack trace for.
        if not os.path.exists(module_path):
            print(f"Error: {module_path} not found\n", file=sys.stderr)
            sys.exit(1)
        if not os.path.isfile(module_path):
            print(f"Error: {module_path} is not a file\n", file=sys.stderr)
            sys.exit(1)
        dirname, filename = os.path.split(module_path)
        module = filename[:-3] if filename.endswith(".py") else filename
        app_dir = dirname

    return f"{module}:{attr}", app_dir


def try_import_module(module: str) -> Optional[types.ModuleType]:
    try:
        if not importlib.util.find_spec(module):
            return None
    except ModuleNotFoundError:
        # find_spec throws this when the module contains both '/' and '.' characters
        return None
    except ImportError:
        # find_spec throws this when the module starts with "."
        return None

    # It's important for ModuleNotFoundError and ImportError (and any other error) NOT
    # to be caught here, as we want to report the true error to the user. Otherwise,
    # missing dependencies can be misreported to the user as the app module itself not
    # being found.
    return importlib.import_module(module)


@main.group(help="""Add files to enhance your Shiny app.""")
def add() -> None:
    pass


@add.command(
    help="""Add a test file for a specified Shiny app.

Add an empty test file for a specified app. You will be prompted with a destination
folder. If you don't provide a destination folder, it will be added in the current
working directory based on the app name.

After creating the shiny app file, you can use `pytest` to run the tests:

        pytest TEST_FILE
"""
)
@click.option(
    "--app",
    "-a",
    type=str,
    help="Please provide the path to the app file for which you want to create a test file.",
)
@click.option(
    "--test-file",
    "-t",
    type=str,
    help="Please provide the name of the test file you want to create. The basename of the test file should start with `test_` and be unique across all test files.",
)
# Param for app.py, param for test_name
def test(
    app: Path | None,
    test_file: Path | None,
) -> None:
    from ._main_add_test import add_test_file

    add_test_file(app_file=app, test_file=test_file)


@main.command(
    help="""Create a Shiny application from a template.

Create an app based on a template. You will be prompted with
a number of application types, as well as the destination folder.
If you don't provide a destination folder, it will be created in the current working
directory based on the template name.

After creating the application, you use `shiny run`:

    shiny run APPDIR/app.py --reload
"""
)
@click.option(
    "--template",
    "-t",
    type=click.STRING,
    help="Choose a template for your new application.",
)
@click.option(
    "--mode",
    "-m",
    type=click.Choice(
        ["core", "express"],
        case_sensitive=False,
    ),
    help="Do you want to use a Shiny Express template or a Shiny Core template?",
)
@click.option(
    "--github",
    "-g",
    help="""
    The GitHub repo containing the template, e.g. 'posit-dev/py-shiny-templates'.
    Can be in the format '{repo_owner}/{repo_name}', '{repo_owner}/{repo_name}@{ref}',
    or '{repo_owner}/{repo_name}:{path}@{ref}'.
    Alternatively, a GitHub URL of the template sub-directory, e.g
    'https://github.com/posit-dev/py-shiny-templates/tree/main/dashboard'.
    """,
)
@click.option(
    "--dir",
    "-d",
    type=str,
    help="The destination directory, you will be prompted if this is not provided.",
)
@click.option(
    "--package-name",
    help="""
    If you are using one of the JavaScript component templates,
    you can use this flag to specify the name of the resulting package without being prompted.
    """,
)
def create(
    template: Optional[str] = None,
    mode: Optional[str] = None,
    github: Optional[str] = None,
    dir: Optional[Path | str] = None,
    package_name: Optional[str] = None,
) -> None:
    from ._main_create import use_github_template, use_internal_template

    if dir is not None:
        dir = Path(dir)

    if github is not None:
        use_github_template(
            github,
            template_name=template,
            mode=mode,
            dest_dir=dir,
            package_name=package_name,
        )
    else:
        use_internal_template(template, mode, dir, package_name)


@main.command(
    help="""The functionality from `shiny static` has been moved to the shinylive package.
Please install shinylive and use `shinylive export` instead of `shiny static`:

  \b
  shiny static-assets remove
  pip install shinylive
  shinylive export APPDIR DESTDIR

""",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
)
def static() -> None:
    print(
        """The functionality from `shiny static` has been moved to the shinylive package.
Please install shinylive and use `shinylive export` instead of `shiny static`:

  shiny static-assets remove
  pip install shinylive
  shinylive export APPDIR DESTDIR
"""
    )
    sys.exit(1)


@main.command(
    no_args_is_help=True,
    help="""Manage local copy of assets for static app deployment. (Deprecated)

    \b
    Commands:
        remove: Remove local copies of assets.
        info: Print information about the local assets.

""",
)
@click.argument("command", type=str)
def static_assets(command: str) -> None:
    dir = _static.get_default_shinylive_dir()

    if command == "remove":
        print(f"Removing {dir}")
        _static.remove_shinylive_local(shinylive_dir=dir)
    elif command == "info":
        _static.print_shinylive_local_info()
    else:
        raise click.UsageError(f"Unknown command: {command}")


@main.command(help="""Convert a JSON file with code cells to a py file.""")
@click.argument(
    "json_file",
    type=str,
)
@click.argument(
    "py_file",
    type=str,
)
def cells_to_app(json_file: str, py_file: str) -> None:
    shiny.quarto.convert_code_cells_to_app_py(json_file, py_file)


@main.command(help="""Get Shiny's HTML dependencies as JSON.""")
def get_shiny_deps() -> None:
    print(shiny.quarto.get_shiny_deps())


class ReloadArgs(TypedDict):
    reload: NotRequired[bool]
    reload_includes: NotRequired[list[str]]
    reload_excludes: NotRequired[list[str]]
    reload_dirs: NotRequired[list[str]]


# Check that the version of rsconnect supports Shiny Express; can be removed in the
# future once this version of rsconnect is widely used. The dependency on "packaging"
# can also be removed then, because it is only used here. (Added 2024-03)
def _verify_rsconnect_version() -> None:
    PACKAGE_NAME = "rsconnect-python"
    MIN_VERSION = "1.22.0"

    from importlib.metadata import PackageNotFoundError, version

    from packaging.version import parse

    try:
        installed_version = parse(version(PACKAGE_NAME))
        required_version = parse(MIN_VERSION)
        if installed_version < required_version:
            print(
                f"Warning: rsconnect-python {installed_version} is installed, but it does not support deploying Shiny Express applications. "
                f"Please upgrade to at least version {MIN_VERSION}. "
                "If you are using pip, you can run `pip install --upgrade rsconnect-python`",
                file=sys.stderr,
            )
    except PackageNotFoundError:
        pass



from __future__ import annotations

import ast
import types
from typing import Any, Callable, cast

from ._helpers import ast_matches_func

sys_alias = "__auto_displayhook_sys__"
expressify_decorator_func_name = "_expressify_decorator_function_def"


class TopLevelTransformer(ast.NodeTransformer):
    def generic_visit(self, node: ast.AST) -> ast.AST:
        return node

    def descend(self, node: ast.AST) -> object:
        return super().generic_visit(node)

    # These are all node types we need to descend into. Any nodes not explicitly
    # sent to super().generic_visit() will go to self.generic_visit(), i.e. the
    # node and its descendants will be left unchanged.

    def visit_Module(self, node: ast.Module) -> object:
        return self.descend(node)

    def visit_With(self, node: ast.With) -> object:
        return self.descend(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> object:
        return self.descend(node)

    def visit_If(self, node: ast.If) -> object:
        return self.descend(node)

    def visit_IfExp(self, node: ast.IfExp) -> object:
        return self.descend(node)

    def visit_For(self, node: ast.For) -> object:
        return self.descend(node)

    def visit_AsyncFor(self, node: ast.AsyncFor) -> object:
        return self.descend(node)

    def visit_While(self, node: ast.While) -> object:
        return self.descend(node)

    def visit_Try(self, node: ast.Try) -> object:
        return self.descend(node)


class FuncBodyDisplayHookTransformer(TopLevelTransformer):
    """
    A NodeTransformer that wraps expressions in a function body's top level and
    within specific flow control statements with sys.displayhook calls, without
    affecting inner functions.

    The NodeTransformer is indiscriminate in what it wraps--the first
    FunctionDef it finds will be transformed.
    """

    def __init__(self, has_docstring: bool = False):
        self.saw_func = False
        self.has_docstring = has_docstring
        self.has_visited_first_node = False

    def visit_FunctionDef(self, node: ast.FunctionDef) -> object:
        """
        Descend into the first (and ONLY the first) function we find.
        """
        if self.saw_func:
            return node

        self.saw_func = True
        return self.descend(node)

    def visit_Expr(self, node: ast.Expr) -> object:
        """
        Each expr we come across will be wrapped in a call to sys.displayhook.
        Note that we don't descend into the node, so child expressions will not
        be affected.
        """
        if not self.has_visited_first_node:
            self.has_visited_first_node = True

            # If the first node is meant to be treated as a docstring, first make sure
            # it actually is a static string, and return it without wrapping it with the
            # displayhook.
            if (
                self.has_docstring
                and isinstance(node.value, ast.Constant)
                and isinstance(node.value.value, str)
            ):
                return node

        return ast.Expr(
            value=ast.Call(
                func=ast.Name(id=sys_alias, ctx=ast.Load()),
                args=[node.value],
                keywords=[],
            )
        )


class TargetFunctionTransformer(ast.NodeTransformer):
    """
    Find the first FunctionDef that matches a target function, and transform it
    using a given function.
    """

    def __init__(
        self, target: types.FunctionType, transform: Callable[[ast.AST], ast.AST]
    ):
        self.target = target
        self.transform = transform
        self.found: ast.FunctionDef | None = None

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        if not self.found and ast_matches_func(node, self.target):
            self.found = node
            new_node = self.transform(node)
            self.found = cast(ast.FunctionDef, new_node)
            return new_node
        else:
            return self.generic_visit(node)


class DisplayFuncsTransformer(TopLevelTransformer):
    # Visit top-level functions and async functions, inserting @sys.displayhook at the
    # top of their decorator lists

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        node.decorator_list.insert(
            0, ast.Name(id=expressify_decorator_func_name, ctx=ast.Load())
        )
        return node

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> object:
        node.decorator_list.insert(
            0, ast.Name(id=expressify_decorator_func_name, ctx=ast.Load())
        )
        return node



import re

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("output_plot")


def test_output_plot_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    plot = controller.OutputPlot(page, "p")

    plot.expect_inline(False)
    plot.expect_height("400px")
    plot.expect_width("100%")

    plot.expect_img_src(re.compile(r"data:image/png;base64"))
    plot.expect_img_height("100%")
    plot.expect_img_width("100%")
    plot.expect_img_alt(None)



from shared import mtcars

from shiny import reactive
from shiny.express import input, render, ui

data = reactive.value(mtcars.iloc[:, range(4)])
with ui.card():
    with ui.layout_column_wrap(fill=False):
        ui.input_action_button("btn", "Filter on columns 0, 1, and 3")
        ui.input_action_button("reset", "Reset column filters")

    @render.data_frame
    def df():
        return render.DataGrid(data(), filters=True)


@reactive.effect
@reactive.event(input.reset)
async def _():
    await df.update_filter(None)


@reactive.effect
@reactive.event(input.btn)
async def _():
    await df.update_filter(
        [
            {"col": 0, "value": [19, 25]},
            {"col": 1, "value": [None, 6]},
            {"col": 3, "value": [100, None]},
        ]
    )



from htmltools import css

from shiny.ui import input_dark_mode


def test_input_dark_mode_style():
    base = input_dark_mode()
    base_style = base.attrs["style"]
    assert isinstance(base_style, str)

    dark_mode = input_dark_mode(style="color: red;")
    assert dark_mode.attrs["style"] == base_style + " color: red;"

    css_position = css(position="absolute", top="1em", left="1em")
    assert isinstance(css_position, str)

    dark_mode = input_dark_mode(style=css_position)
    assert dark_mode.attrs["style"] == base_style + " " + css_position



from __future__ import annotations

__all__ = (
    "input_slider",
    "SliderValueArg",
    "SliderStepArg",
    "AnimationOptions",
)

import math
from datetime import date, datetime, timedelta
from typing import Iterable, Optional, TypeVar, Union, cast

from htmltools import HTML, Tag, TagAttrValue, TagChild, css, div, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from .._typing_extensions import NotRequired, TypedDict
from ._html_deps_external import ionrangeslider_deps
from ._utils import shiny_input_label

# TODO: validate value(s) are within (min,max)?

SliderValueArg = TypeVar("SliderValueArg", float, datetime, date)
SliderStepArg = Union[float, timedelta]


class AnimationOptions(TypedDict):
    """
    Options for the animation of a :func:`input_slider`.

    Parameters
    ----------
    interval
        The interval, in milliseconds, between each animation step.
    loop
        ``True`` to automatically restart the animation when it reaches the end.
    play_button
        Play button text or HTML.
    pause_button
        Pause button text or HTML.

    Returns
    -------
    :
        A TypedDict.

    See Also
    --------
    * :func:`~shiny.ui.input_slider`
    """

    interval: NotRequired[int]
    loop: NotRequired[bool]
    play_button: NotRequired[TagChild]
    pause_button: NotRequired[TagChild]


@add_example()
def input_slider(
    id: str,
    label: TagChild,
    min: SliderValueArg,
    max: SliderValueArg,
    value: SliderValueArg | Iterable[SliderValueArg],
    *,
    step: Optional[SliderStepArg] = None,
    ticks: bool = False,
    animate: bool | AnimationOptions = False,
    width: Optional[str] = None,
    sep: str = ",",
    pre: Optional[str] = None,
    post: Optional[str] = None,
    time_format: Optional[str] = None,
    timezone: Optional[str] = None,
    drag_range: bool = True,
) -> Tag:
    """
    Constructs a slider widget to select a number, date, or date-time from a range.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    min
        The minimum allowed value.
    max
        The maximum allowed value.
    value
        Initial value.
    step
        Interval to use when stepping between min and max.
    ticks
        ``False`` to hide tick marks, ``True`` to show them according to some simple
        heuristics.
    animate
        ``True`` to show simple animation controls with default settings; ``False`` not
        to; or a custom settings list, such as those created using
        :class:`~AnimationOptions`.
    width
        The CSS width, e.g. '400px', or '100%'
    sep
        Separator between thousands places in numbers.
    pre
        A prefix string to put in front of the value.
    post
        A suffix string to put after the value.
    time_format
        Only used if the slider values are :class:`~datetime.date` or
        :class:`~datetime.datetime` objects. A time format string, to be passed to the
        Javascript strftime library. See https://github.com/samsonjs/strftime for more
        details. For Dates, the default is "%F" (like "2015-07-01"), and for Datetimes,
        the default is "%F %T" (like "2015-07-01 15:32:10").
    timezone
        Only used if the values are :class:`~datetime.datetime` objects. A string
        specifying the time zone offset for the displayed times, in the format "+HHMM"
        or "-HHMM". If ``None`` (the default), times will be displayed in the browser's
        time zone. The value "+0000" will result in UTC time.
    drag_range
        This option is used only if it is a range slider (with two values). If ``True``
        (the default), the range can be dragged. In other words, the min and max can be
        dragged together. If ``False``, the range cannot be dragged.

    Returns
    -------
    :
        A UI element

    Notes
    ------
    ::: {.callout-note title="Server value"}
    A number, date, or date-time (depending on the class of value), or in the case of
    slider range, a tuple of two numbers/dates/date-times.
    :::

    See Also
    --------
    * :func:`~shiny.ui.update_slider`
    """

    # Thanks to generic typing, max, value, etc. should be of the same type
    data_type = _slider_type(min)

    # Make sure min, max, value, and step are all numeric
    # (converts dates/datetimes to milliseconds since epoch...this is the value JS wants)
    min_num = _as_numeric(min)
    max_num = _as_numeric(max)
    val_nums = (
        (_as_numeric(value[0]), _as_numeric(value[1]))
        if isinstance(value, (tuple, list))
        else (
            _as_numeric(cast(SliderValueArg, value)),
            _as_numeric(cast(SliderValueArg, value)),
        )
    )
    step_num = _find_step_size(min_num, max_num) if step is None else _as_numeric(step)

    n_ticks = None
    if ticks:
        n_steps = (max_num - min_num) / step_num
        # Make sure there are <= 10 steps.
        # n_ticks can be a noninteger, which is good when the range is not an
        # integer multiple of the step size, e.g., min=1, max=10, step=4
        scale_factor = math.ceil(n_steps / 10)
        n_ticks = n_steps / scale_factor

    id = resolve_id(id)

    props: dict[str, TagAttrValue] = {
        "class": "js-range-slider",
        "id": id,
        "data_skin": "shiny",
        # TODO: do we need to worry about scientific notation (i.e., formatNoSci()?)
        "data_min": str(min_num),
        "data_max": str(max_num),
        "data_from": str(val_nums[0]),
        "data_step": str(step_num),
        "data_grid": ticks,
        "data_grid_num": n_ticks,
        "data_grid_snap": "false",
        "data_prettify_separator": sep,
        "data_prettify_enabled": sep != "",
        "data_prefix": pre,
        "data_postfix": post,
        "data_keyboard": "true",
        "data_data_type": data_type,
        "data_time_format": time_format,
        "data_timezone": timezone,
    }

    if isinstance(value, (tuple, list)):
        props["data_type"] = "double"
        props["data_to"] = str(val_nums[1])
        props["data_drag_interval"] = drag_range

    if not time_format and data_type[0:4] == "date":
        props["data_time_format"] = "%F" if data_type == "date" else "%F %T"

    # ionRangeSlider wants attr = 'true'/'false'
    props = {k: str(v).lower() if isinstance(v, bool) else v for k, v in props.items()}

    resolved_id = resolve_id(id)
    slider_tag = div(
        shiny_input_label(resolved_id, label),
        tags.input(**props),
        *ionrangeslider_deps(),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )

    if animate is False:
        return slider_tag

    if animate is True:
        animate = AnimationOptions()

    animate_tag = div(
        tags.a(
            tags.span(animate.get("play_button", _play_icon()), class_="play"),
            tags.span(animate.get("pause_button", _pause_icon()), class_="pause"),
            href="#",
            class_="slider-animate-button link-secondary",
            data_target_id=resolved_id,
            data_interval=animate.get("interval", 500),
            data_loop=animate.get("loop", True),
        ),
        class_="slider-animate-container",
    )

    slider_tag.append(animate_tag)

    return slider_tag


def _slider_type(x: SliderValueArg) -> str:
    if isinstance(x, datetime):
        return "datetime"
    if isinstance(x, date):
        return "date"
    return "number"


def _as_numeric(x: SliderStepArg | datetime | date) -> float:
    if isinstance(x, timedelta):
        return x.total_seconds() * 1000
    if isinstance(x, datetime):
        return x.timestamp() * 1000
    if isinstance(x, date):
        return datetime(x.year, x.month, x.day).timestamp() * 1000
    return x


def _find_step_size(min: int | float, max: int | float) -> int | float:
    # TODO: this is a naive version of shiny::findStepSize() that might be susceptible to
    # rounding errors? https://github.com/rstudio/shiny/pull/1956
    range = max - min

    if range < 2 or isinstance(min, float) or isinstance(max, float):
        step = range / 100
        # Round the step to get rid of any floating point arithmetic errors by
        # mimicing what signif(digits = 10, step) does in R (see Description of ?signif)
        # (the basic intuition is that smaller differences need more precision)
        return round(step, 10 - math.ceil(math.log10(step)))
    else:
        return 1


play_icon = """<svg viewBox="0 0 448 512" preserveAspectRatio="none" aria-hidden="true" role="img" style="fill:currentColor;height:1em;width:0.88em;margin-left:auto;margin-right:0.2em;position:relative;vertical-align:-0.125em;font-size:inherit;overflow:visible;">
  <title>Play</title>
  <!-- Font Awesome Free 5.15.4 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License) -->
  <path d="M424.4 214.7L72.4 6.6C43.8-10.3 0 6.1 0 47.9V464c0 37.5 40.7 60.1 72.4 41.3l352-208c31.4-18.5 31.5-64.1 0-82.6z"></path>
</svg>"""

pause_icon = """<svg viewBox="0 0 448 512" preserveAspectRatio="none" aria-hidden="true" role="img" style="fill:currentColor;height:1em;width:0.88em;margin-left:auto;margin-right:0.2em;position:relative;vertical-align:-0.125em;font-size:inherit;overflow:visible;">
  <title>Pause</title>
  <!-- Font Awesome Free 5.15.4 by @fontawesome - https://fontawesome.com License - https://fontawesome.com/license/free (Icons: CC BY 4.0, Fonts: SIL OFL 1.1, Code: MIT License) -->
  <path d="M144 479H48c-26.5 0-48-21.5-48-48V79c0-26.5 21.5-48 48-48h96c26.5 0 48 21.5 48 48v352c0 26.5-21.5 48-48 48zm304-48V79c0-26.5-21.5-48-48-48h-96c-26.5 0-48 21.5-48 48v352c0 26.5 21.5 48 48 48h96c26.5 0 48-21.5 48-48z"></path>
</svg>"""


def _play_icon() -> Tag | HTML:
    return HTML(play_icon)


def _pause_icon() -> Tag | HTML:
    return HTML(pause_icon)



from shiny import App, Inputs, ui

app_ui = ui.page_fluid(
    ui.navset_bar(
        ui.nav_panel("Page 1", "Page 1 content"),
        ui.nav_panel(
            "Page 2",
            ui.navset_card_underline(
                ui.nav_panel("Tab 1", "Tab 1 content"),
                ui.nav_panel("Tab 2", "Tab 2 content"),
                ui.nav_panel("Tab 3", "Tab 3 content"),
            ),
        ),
        title="Nav Panel Example",
    ),
)


def server(input: Inputs):
    pass


app = App(app_ui, server)



from __future__ import annotations

from typing import Generic

from .._deprecated import ShinyDeprecationWarning
from .transformer._transformer import IT, OT, ValueFnAsync, ValueFnSync

# ======================================================================================
# Deprecated classes
# ======================================================================================


# A RenderFunction object is given a app-supplied function which returns an `IT`. When
# the .__call__ method is invoked, it calls the app-supplied function (which returns an
# `IT`), then converts the `IT` to an `OT`. Note that in many cases but not all, `IT`
# and `OT` will be the same.
class RenderFunction(Generic[IT, OT]):
    """
    Deprecated. Please use :class:`~shiny.render.renderer.Renderer` class instead.
    """

    def __init__(self, fn: ValueFnSync[IT]) -> None:
        raise ShinyDeprecationWarning(
            "Class `"
            + str(self.__class__.__name__)
            + "` inherits from the deprecated class `shiny.render.RenderFunction`. "
            "Please update your renderer to use `shiny.render.renderer.Renderer` instead."
        )


# The reason for having a separate RenderFunctionAsync class is because the __call__
# method is marked here as async; you can't have a single class where one method could
# be either sync or async.
class RenderFunctionAsync(Generic[IT, OT]):
    """
    Deprecated. Please use :class:`~shiny.render.renderer.Renderer` class instead.
    """

    def __init__(self, fn: ValueFnAsync[IT]) -> None:
        raise ShinyDeprecationWarning(
            "Class `"
            + str(self.__class__.__name__)
            + "` inherits from the deprecated class `shiny.render.RenderFunctionAsync`. "
            "Please update your renderer to use `shiny.render.renderer.Renderer` instead."
        )



import pathlib

import pandas as pd

from shiny import App, Inputs, Outputs, Session, render, ui

dir = pathlib.Path(__file__).parent
mtcars = pd.read_csv(dir / "mtcars.csv")


app_ui = ui.page_fluid(
    ui.input_checkbox("highlight", "Highlight min/max values"),
    ui.output_table("result"),
    # Legend
    ui.panel_conditional(
        "input.highlight",
        ui.panel_absolute(
            "Yellow is maximum, grey is minimum",
            bottom="6px",
            right="6px",
            class_="p-1 bg-light border",
        ),
    ),
    class_="p-3",
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.table
    def result():
        if not input.highlight():
            # If we're not highlighting values, we can simply
            # return the pandas data frame as-is; @render.table
            # will call .to_html() on it.
            return mtcars
        else:
            # We need to use the pandas Styler API. The default
            # formatting options for Styler are not the same as
            # DataFrame.to_html(), so we set a few options to
            # make them match.
            return (
                mtcars.style.set_table_attributes(
                    'class="dataframe shiny-table table w-auto"'
                )
                .hide(axis="index")
                .format(
                    {
                        "mpg": "{0:0.1f}",
                        "disp": "{0:0.1f}",
                        "drat": "{0:0.2f}",
                        "wt": "{0:0.3f}",
                        "qsec": "{0:0.2f}",
                    }
                )
                .set_table_styles(
                    [dict(selector="th", props=[("text-align", "right")])]
                )
                .highlight_min(color="silver")
                .highlight_max(color="yellow")
            )


app = App(app_ui, server)



from shiny.express import ui

ui.page_opts(
    id="page_navbar_fillable",
    fillable=True,
    gap="300px",
)

with ui.nav_panel("Data"):
    "This page could be used to pick a dataset."

with ui.nav_panel("View"):
    "This page could be used to view the dataset."



from shiny import reactive
from shiny.express import input, ui

"The first checkbox group controls the second"
ui.input_checkbox_group(
    "inCheckboxGroup", "Input checkbox", ["Item A", "Item B", "Item C"]
)
ui.input_checkbox_group(
    "inCheckboxGroup2", "Input checkbox 2", ["Item A", "Item B", "Item C"]
)


@reactive.effect
def _():
    x = input.inCheckboxGroup()

    # Can also set the label and select items
    ui.update_checkbox_group(
        "inCheckboxGroup2",
        label="Checkboxgroup label " + str(len(x)),
        choices=x,
        selected=x,
    )



from shiny import reactive
from shiny.express import input, ui

ui.input_action_button("add", "Add UI")


@reactive.effect
@reactive.event(input.add)
def _():
    ui.insert_ui(
        ui.input_text("txt" + str(input.add()), "Enter some text"),
        selector="#add",
        where="afterEnd",
    )



from shiny import reactive
from shiny.express import input, ui

with ui.layout_column_wrap(width=1 / 2):
    ui.input_radio_buttons("pet_type", "Pet type", ["Dog", "Cat", "Bird"], inline=True)
    ui.input_radio_buttons("pet_sex", "Pet sex", ["Male", "Female"], inline=True)
    ui.input_text("name", "Pet name", "Charlie")
    ui.input_text("royal_name", "Royal Name", "King Charlie")


@reactive.effect
@reactive.event(input.pet_type)
def _():
    # Update the label of the pet name input
    ui.update_text("name", label=f"{input.pet_type()}'s name")


@reactive.effect
def _():
    # Update the value of the royal name input
    royal_noun = "King" if input.pet_sex() == "Male" else "Queen"
    ui.update_text("royal_name", value=f"{royal_noun} {input.name()}")



from shiny import App, reactive, ui

app_ui = ui.page_fixed(
    ui.input_slider(
        "receiver", "Receiver:", min=0, max=100, value=50, step=1, width="100%"
    ),
    ui.p(
        "Change the min and max values below to see the receiver slider above update."
    ),
    ui.layout_column_wrap(
        ui.input_slider("min", "Min:", min=0, max=50, value=0, step=1),
        ui.input_slider("max", "Max:", min=50, max=100, value=100, step=1),
        width=1 / 2,
    ),
)


def server(input, output, session):
    @reactive.effect
    def _():
        # You can update the value, min, max, and step.
        ui.update_slider(
            "receiver",
            value=max(min(input.receiver(), input.max()), input.min()),
            min=input.min(),
            max=input.max(),
        )


app = App(app_ui, server)



import os
from urllib.parse import urlparse

import pytest
from examples.example_apps import reruns, reruns_delay
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def get_spinner_computed_property(
    page: Page, element_id: str, property_name: str
) -> str:
    expect(page.locator(element_id)).to_be_visible()
    return page.evaluate(
        f"window.getComputedStyle(document.querySelector('{element_id}'), '::after').getPropertyValue('{property_name}');"
    )


def get_pulse_computed_property(page: Page, property_name: str) -> str:
    expect(page.locator("html body")).to_be_visible()
    return page.evaluate(
        f"window.getComputedStyle(document.documentElement, '::after').getPropertyValue('{property_name}');"
    )


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_busy_indicators(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)
    spinner_type = controller.InputRadioButtons(page, "busy_indicator_type")
    render_button = controller.InputTaskButton(page, "rerender")

    # Verify spinner indicator behavior
    spinner_properties = [
        ("#pulse-plot", "50px", "rgb(128, 128, 0)", "pulse.svg"),
        ("#ring-plot", "10px", "rgb(255, 0, 0)", "ring.svg"),
        ("#bars-plot", "20px", "rgb(0, 128, 0)", "bars.svg"),
        ("#dots-plot", "30px", "rgb(0, 0, 255)", "dots.svg"),
    ]

    for element_id, height, background_color, svg_name in spinner_properties:
        assert get_spinner_computed_property(page, element_id, "height") == height
        assert (
            get_spinner_computed_property(page, element_id, "background-color")
            == background_color
        )
        mask_image_url = get_spinner_computed_property(page, element_id, "mask-image")
        assert os.path.basename(urlparse(mask_image_url).path).rstrip('")') == svg_name
        assert get_spinner_computed_property(page, element_id, "width") == height

    # Verify pulse indicator behavior
    output_txt = controller.OutputTextVerbatim(page, "counter")
    output_txt.expect_value("0")
    spinner_type.set("pulse")
    render_button.click()
    # `::after` is not an implemented selector in playwright
    # Since we are not using locators wait for up to 2 secs
    for _ in range(200):
        if get_pulse_computed_property(page, "height") != "auto":
            break
        page.wait_for_timeout(10)
    assert get_pulse_computed_property(page, "height") == "100px"
    assert (
        get_pulse_computed_property(page, "background-image")
        == "linear-gradient(45deg, rgb(0, 0, 255), rgb(255, 0, 0))"
    )
    output_txt.expect_value("1")
    # since output value is 1, nothing is computing
    # verify the pulse indicator is removed
    assert get_pulse_computed_property(page, "height") == "auto"



import duckdb

from shiny import module, reactive, render, ui


@module.ui
def query_output_ui(remove_id, qry="SELECT * from weather LIMIT 10"):
    out = ui.row(
        {
            "style": "border: 1px solid gray; border-radius: 5px; margin:10px",
            "id": remove_id,
        },
        ui.column(
            3,
            ui.div(
                {"style": "font-family: var(--bs-font-monospace); margin-top: 10px;"},
                ui.input_text_area(
                    "sql_query",
                    "",
                    value=qry,
                    width="100%",
                    height="200px",
                ),
            ),
            ui.row(
                {"style": "margin-bottom: 10px;"},
                ui.column(
                    6,
                    ui.input_action_button("run", "Run Query", width="90%"),
                ),
                ui.column(
                    6,
                    ui.input_action_button("rmv", "Remove Query", width="90%"),
                ),
            ),
        ),
        ui.column(9, ui.output_data_frame("results")),
    )

    return out


@module.server
def query_output_server(
    input, output, session, con: duckdb.DuckDBPyConnection, remove_id
):
    @render.data_frame
    def results():
        # In order to avoid the query re-running with each keystroke we
        # wrap it in isolate and add a call to `input.run()` to trigger execution.
        # This ensures that the query runs when the module is populated and also
        # passes the error messages on to the user.
        input.run()
        with reactive.isolate():
            qry = input.sql_query().replace("\n", " ")
            result = con.query(qry).to_df()

        return result

    @reactive.effect
    @reactive.event(input.rmv)
    def _():
        ui.remove_ui(selector=f"div#{remove_id}")



# TODO-barret; Schedule meeting with Joe when done

import inspect

# from types import UnboundMethodType
# import types
from functools import wraps
from typing import Any, Callable, TypeVar
from weakref import WeakKeyDictionary

from ... import reactive

R = TypeVar("R")
S = TypeVar("S")

# TODO-future: Can we do this for effect?
# reactive_value_method
#     @reactive.value
#     def __some_effect(self):
#         # Should be possible as it doesn't need to run on init, only on demand!
#         ...
# reactive_effect_method
#     @reactive.effect
#     def __some_effect(self):
#         # Thinking it is impossible to do!
#         # There is no way to get the `session` within the init method
#         # We should throw an error if we find this situation
#         ...


def is_class_method(fn: Callable[..., Any]) -> bool:
    """
    Determine if a function is a class method.

    This function checks if a function is a class method by checking if the function's
    `__qualname__` ends with a non-`<locals>` value before the function name.

    Parameters
    ----------
    fn
        The function to check.

    Returns
    -------
    :
        True if the function is a class method, False otherwise.
    """
    # if not inspect.isfunction(fn) and not inspect.ismethod(fn):
    if not inspect.isfunction(fn):
        return False

    # Check if the function's __qualname__ contains a class name
    if "." not in fn.__qualname__:
        # If no `.` in the qualname, it is not a class method
        return False

    # `__qualname__` is the fully qualified name of the function
    # It contains the class name if the function is a class method
    # If the qualname has `<locals>` just before the final `__name__` value, it is a nested function.
    # (Nested Class definitions are ok. Nested function definitions are not)
    scope_name, fn_name = fn.__qualname__.rsplit(".", 1)

    # If we find a nested function, it is not a class method
    if scope_name.endswith("<locals>"):
        return False

    if fn_name != fn.__name__:
        # If the final value of the qualname should always be the function name
        return False

    # Must be a class method!
    return True


def reactive_calc_method(fn: Callable[[S], R]) -> Callable[[S], R]:

    # Can not use `inspect.ismethod` as it will return False for class definition methods,
    # but True for class instance methods. Ex: Barret.get_name vs barret.get_name

    if not (inspect.isfunction(fn) and is_class_method(fn)):
        raise TypeError("reactive_calc_method should only be used on class methods")

    calc_cache: WeakKeyDictionary[S, reactive.Calc_[R]] = WeakKeyDictionary()

    @wraps(fn)
    def _(self: S) -> R:

        if self not in calc_cache:

            # # While the code below is more concise, functools.partial does not create
            # # `.__docs__` and `.__name__` attributes
            # # https://docs.python.org/3/library/functools.html#partial-objects
            # calc_cache[self] = reactive.calc(functools.partial(fn, self))

            @reactive.calc
            @wraps(fn)
            def calc_fn():
                return fn(self)

            # Set calc method
            calc_cache[self] = calc_fn

            # Store the reactive calc function on self's fn, so that `calc_fn` does not
            # get garbage collected. Before these lines, it is isolated under a weak
            # key. After these lines, it is stored on self's `__dict__` and should not
            # be garbage collected until `self` is garbage collected.
            if not hasattr(self.__dict__, "_reactive_calc_method"):
                self.__dict__["_reactive_calc_method"] = {}
            self_reactive_calc_cache: dict[str, reactive.Calc_[R]] = self.__dict__[
                "_reactive_calc_method"
            ]
            if hasattr(self_reactive_calc_cache, fn.__name__):
                raise AttributeError(
                    f"Reactive calc method `{fn.__name__}` has already be cached on self: {self}"
                )
            self_reactive_calc_cache[fn.__name__] = calc_fn

        return calc_cache[self]()

    return _



from icons import gear_fill

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.popover(
        ui.input_action_button("btn", "A button", class_="mt-3"),
        "A popover with more context and information than should be used in a tooltip.",
        "You can even have multiple DOM elements in a popover!",
        id="btn_popover",
    ),
    ui.hr(),
    ui.card(
        ui.card_header(
            "Plot title (Click the gear to change variables)",
            ui.popover(
                ui.span(
                    gear_fill,
                    style="position:absolute; top: 5px; right: 7px;",
                ),
                "Put dropdowns here to alter your plot!",
                ui.input_selectize("x", "X", ["x1", "x2", "x3"]),
                ui.input_selectize("y", "Y", ["y1", "y2", "y3"]),
                placement="right",
                id="card_popover",
            ),
        ),
        ui.output_text_verbatim("plot_txt", placeholder=True),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def plot_txt():
        return f"<Making plot using x: {input.x()} and y: {input.y()}>"


app = App(app_ui, server=server)



from shiny import reactive
from shiny.express import input, render, ui

ui.input_text("name", "Your Name")
ui.input_action_button("greet", "Say Hello", disabled=True)


@reactive.effect
@reactive.event(input.name)
def set_button_state():
    if input.name():
        ui.update_action_button("greet", disabled=False)
    else:
        ui.update_action_button("greet", disabled=True)


@render.ui
@reactive.event(input.greet)
def hello():
    return ui.p(f"Hello, {input.name()}!", class_="fs-1 text-primary mt-3")



import asyncio
from datetime import datetime, timezone

from shiny import reactive, render
from shiny.express import input, ui

ui.h5("Current time")


@render.text()
def current_time() -> str:
    reactive.invalidate_later(0.1)
    return str(datetime.now(timezone.utc).isoformat())


with ui.p():
    "Notice that the time above updates every second, even if you click the button below."


@ui.bind_task_button(button_id="btn_task")
@reactive.extended_task
async def slow_compute(a: int, b: int) -> int:
    await asyncio.sleep(1.5)
    return a + b


async def slow_input_compute(a: int, b: int) -> int:
    await asyncio.sleep(1.5)
    return a + b


with ui.layout_sidebar():
    with ui.sidebar():
        ui.input_numeric("x", "x", 1)
        ui.input_numeric("y", "y", 2)
        ui.input_task_button("btn_task", "Non-blocking task")
        ui.input_task_button("btn_block", "Block compute", label_busy="Blocking...")
        ui.input_action_button("btn_cancel", "Cancel")

    @reactive.effect
    @reactive.event(input.btn_task, ignore_none=False)
    def handle_click():
        # slow_compute.cancel()
        slow_compute(input.x(), input.y())

    @reactive.effect
    @reactive.event(input.btn_block, ignore_none=False)
    async def handle_click2():
        # slow_compute.cancel()
        await slow_input_compute(input.x(), input.y())

    @reactive.effect
    @reactive.event(input.btn_cancel)
    def handle_cancel():
        slow_compute.cancel()

    ui.h5("Sum of x and y")

    @render.text
    def show_result():
        return str(slow_compute.result())



from __future__ import annotations

from pathlib import Path, PurePath
from typing import Literal, Union

import pytest

from .._docstring import no_example
from ..run._run import shiny_app_gen

__all__ = (
    "create_app_fixture",
    "ScopeName",
)


ScopeName = Literal["session", "package", "module", "class", "function"]
"""
Pytest fixture scopes

Fixtures are created when first requested by a test, and are destroyed based on their scope:

* `function`: the default scope, the fixture is destroyed at the end of the test.
* `class`: the fixture is destroyed during teardown of the last test in the class.
* `module`: the fixture is destroyed during teardown of the last test in the module.
* `package`: the fixture is destroyed during teardown of the last test in the package.
* `session`: the fixture is destroyed at the end of the test session.

**Note:** Pytest only caches one instance of a fixture at a time, which means that when using a parametrized fixture, pytest may invoke a fixture more than once in the given scope.

Documentation taken from [https://docs.pytest.org/en/7.1.x/how-to/fixtures.html#fixture-scopes](https://docs.pytest.org/en/7.1.x/how-to/fixtures.html#fixture-scopes)
"""


@no_example()
def create_app_fixture(
    app: Union[PurePath, str],
    scope: ScopeName = "module",
):
    """
    Create a fixture for a local Shiny app directory.

    Creates a fixture for a local Shiny app that is not contained within the same folder. This fixture is used to start the Shiny app process and return the local URL of the app.

    If the app path is located in the same directory as the test file, then `create_app_fixture()` can be skipped and `local_app` test fixture can be used instead.

    Parameters
    ----------
    app
        The path to the Shiny app file.

        If `app` is a `Path` or `PurePath` instance and `Path(app).is_file()` returns
        `True`, then this value will be used directly. Note, `app`'s file path will be
        checked from where corresponding `pytest` test is collected, not necessarily
        where `create_app_fixture()` is called.

        Otherwise, all `app` paths will be considered to be relative paths from where
        the test function was collected.

        To be sure that your `app` path is always relative, supply a `str` value.
    scope
        The scope of the fixture.

    Returns
    -------
    fixture_func
        The fixture function.

    Examples
    --------
    ```python
    from playwright.sync_api import Page

    from shiny.playwright import controller
    from shiny.pytest import create_app_fixture
    from shiny.run import ShinyAppProc

    # The variable name `app` MUST match the parameter name in the test function
    app = create_app_fixture("relative/path/to/app.py")


    def test_app_code(page: Page, app: ShinyAppProc):

        page.goto(app.url)
        # Add test code here
        ...
    ```
    """

    @pytest.fixture(scope=scope)
    def fixture_func(request: pytest.FixtureRequest):
        app_purepath_exists = isinstance(app, PurePath) and Path(app).is_file()
        app_path = app if app_purepath_exists else request.path.parent / app
        sa_gen = shiny_app_gen(app_path)
        yield next(sa_gen)

    return fixture_func



from __future__ import annotations

import datetime
import typing
from typing import Literal

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_date_range")


def expect_date_range(
    date: controller.InputDateRange,
    start_value: str | Literal["today"] = "today",
    end_value: str | Literal["today"] = "today",
    *,
    autoclose: bool = True,
    format: str = "yyyy-mm-dd",
    language: str = "en",
    max_date: typing.Optional[str] = None,
    min_date: typing.Optional[str] = None,
    startview: str = "month",
    weekstart: int = 0,
    width: typing.Optional[str] = None,
    separator: str = " to ",
) -> None:
    start_value = str(datetime.date.today()) if start_value == "today" else start_value
    end_value = str(datetime.date.today()) if end_value == "today" else end_value
    date.expect_value((start_value, end_value))
    autoclose_str = "true" if autoclose else "false"
    date.expect_autoclose(autoclose_str)
    date.expect_format(format)
    date.expect_language(language)
    date.expect_max_date(max_date)
    date.expect_min_date(min_date)
    date.expect_startview(startview)
    date.expect_weekstart(weekstart)
    date.expect_width(width)
    date.expect_separator(separator)


def test_input_date_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    daterange1 = controller.InputDateRange(page, "daterange1")

    daterange1.expect_label("Date range:")
    today = str(datetime.date.today().strftime("%Y-%m-%d"))
    tommorow = str(
        (datetime.date.today() + datetime.timedelta(days=1)).strftime("%Y-%m-%d")
    )
    daterange1.set((today, today))
    expect_date_range(daterange1, today, today)
    daterange1.set((today, tommorow))

    expect_date_range(
        controller.InputDateRange(page, "daterange2"), "2001-01-01", "2010-12-31"
    )

    expect_date_range(
        controller.InputDateRange(page, "daterange3"),
        "01/01/01",
        "12/31/10",
        format="mm/dd/yy",
        separator=" - ",
        min_date="2001-01-01",
        max_date="2012-12-21",
    )

    expect_date_range(
        controller.InputDateRange(page, "daterange4"),
        str(datetime.date(2001, 1, 1)),
        str(datetime.date(2010, 12, 31)),
    )
    expect_date_range(
        controller.InputDateRange(page, "daterange5"),
        language="de",
        weekstart=1,
    )
    expect_date_range(
        controller.InputDateRange(page, "daterange6"),
        startview="decade",
    )

    expect_date_range(
        controller.InputDateRange(page, "daterange7"),
        width="600px",
    )

    expect_date_range(
        controller.InputDateRange(page, "daterange8"),
        autoclose=False,
    )



import os
import time

import numpy as np
import seaborn as sns

from shiny import App, render, ui

app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.input_selectize(
            "indicator_types",
            "Busy indicator types",
            ["spinners", "pulse"],
            multiple=True,
            selected=["spinners", "pulse"],
        ),
        ui.download_button("download", "Download source"),
    ),
    ui.card(
        ui.card_header(
            "Plot that takes a few seconds to render",
            ui.input_task_button("simulate", "Simulate"),
            class_="d-flex justify-content-between align-items-center",
        ),
        ui.output_plot("plot"),
    ),
    ui.busy_indicators.options(spinner_type="bars3"),
    ui.output_ui("indicator_types_ui"),
    title="Busy indicators demo",
)


def server(input):

    @render.plot
    def plot():
        input.simulate()
        time.sleep(3)
        sns.lineplot(x=np.arange(100), y=np.random.randn(100))

    @render.ui
    def indicator_types_ui():
        return ui.busy_indicators.use(
            spinners="spinners" in input.indicator_types(),
            pulse="pulse" in input.indicator_types(),
        )

    @render.download
    def download():
        time.sleep(3)
        path = os.path.join(os.path.dirname(__file__), "app-core.py")
        return path


app = App(app_ui, server)



from __future__ import annotations

from typing import Optional, Protocol

from htmltools import (
    HTML,
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagFunction,
    TagList,
    css,
    div,
    tags,
)

from .._docstring import add_example
from .._namespaces import resolve_id_or_none
from .._utils import private_random_id
from ..types import MISSING, MISSING_TYPE
from ._html_deps_shinyverse import components_dependencies
from ._tag import consolidate_attrs
from ._tooltip import tooltip
from .css._css_unit import CssUnit, as_css_padding, as_css_unit
from .fill import as_fill_item, as_fillable_container

# TODO-future; Update header to return CardHeader class. Same for footer. Then we
# can check `*args` for a CardHeader class and move it to the top. And footer to the
# bottom. Can throw error if multiple headers/footers are provided or could concatenate.


__all__ = (
    "CardItem",
    "card",
    "card_header",
    "card_footer",
)

############################################################################
# Experimental+ full implementation
############################################################################
# Reasons for experimental implementations:
# * Missing parameter of `wrapper=`
# * Missing helper function `card_body()`
############################################################################


@add_example()
def card(
    *args: TagChild | TagAttrs | CardItem,
    full_screen: bool = False,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    id: Optional[str] = None,
    # wrapper: WrapperCallable | None | MISSING_TYPE = MISSING,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    A Bootstrap card component

    A general purpose container for grouping related UI elements together with a border
    and optional padding. To learn more about `card()`s, see [this
    article](https://rstudio.github.io/bslib/articles/cards.html).

    Parameters
    ----------
    *args
        UI elements.
    full_screen
        If `True`, an icon will appear when hovering over the card body. Clicking the
        icon expands the card to fit viewport size.
    height,max_height,min_height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen`.
    fill
        Whether or not to allow the card to grow/shrink to fit a fillable container with
        an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    class_
        Additional CSS classes for the returned Tag.
    id
        Provide a unique identifier for the :func:`~shiny.ui.card` or to report its
        full screen state to Shiny. For example, using `id="my_card"`, you can observe
        the card's full screen state with `input.my_card_full_screen()`.
    **kwargs
        HTML attributes on the returned Tag.

    Returns
    -------
    :
        An :func:`~shiny.ui.tags.div` tag.

    See Also
    --------
    * :func:`~shiny.ui.card_header` for creating a header within the card.
    * :func:`~shiny.ui.card_footer` for creating a footer within the card.
    """
    # * :func:`~shiny.ui.layout_column_wrap` for laying out multiple cards
    #   (or multiple columns inside a card).
    return _card_impl(
        *args,
        full_screen=full_screen,
        height=height,
        max_height=max_height,
        min_height=min_height,
        fill=fill,
        class_=class_,
        id=id,
        wrapper=MISSING,
        **kwargs,
    )


def _card_impl(
    *args: TagChild | TagAttrs | CardItem,
    full_screen: bool = False,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    id: Optional[str] = None,
    wrapper: WrapperCallable | None | MISSING_TYPE = MISSING,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Common implementation method between `shiny.ui.card()` and `shiny.experimental.ui.card()`.

    Experimental implements the parameter `wrapper=`. Main does not as `card_body()` is not in `main`.
    """
    # TODO-question Should card_body() ever exist in main?
    # Barret: Yes? Need to customize fill/fillable content (ex: Fixed height for item B where content = A, B, C; fill = TRUE, fillable = TRUE).
    if isinstance(wrapper, MISSING_TYPE):
        wrapper = card_body

    attrs, children = consolidate_attrs(*args, class_=class_, **kwargs)
    children = _wrap_children_in_card(*children, wrapper=wrapper)

    id = resolve_id_or_none(id)
    is_shiny_input = id is not None

    if full_screen and id is None:
        id = private_random_id("bslib_card")

    tag = div(
        {
            "id": id,
            "class": "card bslib-card bslib-mb-spacing",
            "style": css(
                height=as_css_unit(height),
                max_height=as_css_unit(max_height),
                min_height=as_css_unit(min_height),
            ),
            "data-bslib-card-init": True,
            "data-full-screen": "false" if full_screen else None,
        },
        {"class": "bslib-card-input"} if is_shiny_input else None,
        *children,
        attrs,
        _full_screen_toggle(id) if full_screen else None,
        components_dependencies(),
        _card_js_init(),
    )
    if fill:
        tag = as_fill_item(tag)
    tag = as_fillable_container(tag)

    return tag


def _card_js_init() -> Tag:
    return tags.script(
        {"data-bslib-card-init": True},
        "window.bslib.Card.initializeAllCards();",
    )


def _full_screen_toggle(id_controls: TagAttrValue) -> Tag:
    return tooltip(
        tags.button(
            {
                "class": "bslib-full-screen-enter badge rounded-pill",
                "aria-expanded": "false",
                "aria-controls": id_controls,
                "aria-label": "Expand card",
            },
            _full_screen_toggle_icon(),
        ),
        "Expand",
    )


# via bsicons::bs_icon("arrows-fullscreen")
def _full_screen_toggle_icon() -> HTML:
    # https://www.visiwig.com/icons/
    # https://www.visiwig.com/icons-license/
    return HTML(
        '<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24" style="height:1em;width:1em;fill:currentColor;" aria-hidden="true" role="img"><path d="M20 5C20 4.4 19.6 4 19 4H13C12.4 4 12 3.6 12 3C12 2.4 12.4 2 13 2H21C21.6 2 22 2.4 22 3V11C22 11.6 21.6 12 21 12C20.4 12 20 11.6 20 11V5ZM4 19C4 19.6 4.4 20 5 20H11C11.6 20 12 20.4 12 21C12 21.6 11.6 22 11 22H3C2.4 22 2 21.6 2 21V13C2 12.4 2.4 12 3 12C3.6 12 4 12.4 4 13V19Z"/></svg>'
    )


# https://mypy.readthedocs.io/en/stable/protocols.html#callback-protocols
class WrapperCallable(Protocol):
    """
    A callable that wraps children into a :class:`~shiny.ui.CardItem`.
    """

    def __call__(self, *args: TagChild) -> CardItem:
        """
        Wraps children into a :class:`~shiny.ui.CardItem`.

        Parameters
        ----------
        *args
            `TagChild` children to wrap.

        Returns
        -------
        :
            A :class:`~shiny.ui.CardItem` object.
        """
        ...


def _as_card_items(
    *children: TagChild | CardItem | None,  # `TagAttrs` are not allowed here
    wrapper: WrapperCallable | None,
) -> list[CardItem]:
    # We don't want `None`s creating empty card bodies
    children_vals = [child for child in children if child is not None]

    attrs, children_vals = consolidate_attrs(*children_vals)
    if len(attrs) > 0:
        raise ValueError("`TagAttrs` are not allowed in `_as_card_items(*children=)`.")

    if not callable(wrapper):
        ret: list[CardItem] = []
        for child in children_vals:
            if isinstance(child, CardItem):
                ret.append(child)
            else:
                ret.append(CardItem(child))
        return ret

    # Any children that are `is.card_item` should be included verbatim. Any
    # children that are not, should be wrapped in card_body(). Consecutive children
    # that are not card_item, should be wrapped in a single card_body().
    state = "asis"  # "wrap" | "asis"
    new_children: list[CardItem] = []
    children_to_wrap: list[TagChild] = []

    def wrap_children():
        nonlocal children_to_wrap
        wrapped_children = wrapper(*children_to_wrap)
        new_children.append(wrapped_children)
        children_to_wrap = []

    for child in children_vals:
        if isinstance(child, CardItem):
            if state == "wrap":
                wrap_children()
            state = "asis"
            new_children.append(child)
        else:
            # Not a card, collect it for wrapping
            state = "wrap"
            children_to_wrap.append(child)
    if state == "wrap":
        wrap_children()

    return new_children


def _wrap_children_in_card(
    *children: TagChild | CardItem | None,  # `TagAttrs` are not allowed here
    wrapper: WrapperCallable | None,
) -> list[TagChild]:
    card_items = _as_card_items(*children, wrapper=wrapper)
    tag_children = [card_item.resolve() for card_item in card_items]
    return tag_children


@add_example()
def card_body(
    *args: TagChild | TagAttrs,
    fillable: bool = True,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    max_height_full_screen: Optional[CssUnit] | MISSING_TYPE = MISSING,
    height: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    gap: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> CardItem:
    # For a general overview of the :func:`~shiny.ui.card` API, see [this article](https://rstudio.github.io/bslib/articles/cards.html).
    """
    Card body container

    A general container for the "main content" of a :func:`~shiny.ui.card`. This
    component is designed to be provided as direct children to :func:`~shiny.ui.card`.

    Parameters
    ----------
    *args
        Contents to the card's body. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    fillable
        Whether or not the card item should be a fillable (i.e. flexbox) container.
    min_height,max_height,max_height_full_screen
        Any valid CSS length unit. If `max_height_full_screen` is missing, it is set to
        `max_height`.
    height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
        `full_screen` (in this case, consider setting a `height` in
        `card_body()`).
    padding
        Padding to use for the body. This can be a numeric vector
        (which will be interpreted as pixels) or a character vector with valid CSS
        lengths. The length can be between one and four. If one, then that value
        will be used for all four sides. If two, then the first value will be used
        for the top and bottom, while the second value will be used for left and
        right. If three, then the first will be used for top, the second will be
        left and right, and the third will be bottom. If four, then the values will
        be interpreted as top, right, bottom, and left respectively.
    gap
        A CSS length unit defining the `gap` (i.e., spacing) between elements provided
        to `*args`. This argument is only applicable when `fillable = TRUE`.
    fill
        Whether to allow this element to grow/shrink to fit its `card` container.
    class_
        Additional CSS classes for the returned Tag.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A :class:`~shiny.ui.CardItem` object.

    See Also
    --------
    * :func:`~shiny.ui.layout_column_wrap` for laying out multiple cards
        (or multiple columns inside a card).
    * :func:`~shiny.ui.card` for creating a card component.
    * :func:`~shiny.ui.card_header` for creating a header within the card.
    * :func:`~shiny.ui.card_footer` for creating a footer within the card.
    """
    if isinstance(max_height_full_screen, MISSING_TYPE):
        max_height_full_screen = max_height

    div_style_args = {
        "min-height": as_css_unit(min_height),
        "--bslib-card-body-max-height": as_css_unit(max_height),
        "--bslib-card-body-max-height-full-screen": as_css_unit(max_height_full_screen),
        "margin-top": "auto",
        "margin-bottom": "auto",
        # .card-body already adds `flex: 1 1 auto` so make sure to override it
        "flex": "1 1 auto" if fill else "0 0 auto",
        "padding": as_css_padding(padding),
        "gap": as_css_unit(gap),
        "height": as_css_unit(height),
    }
    tag = tags.div(
        {
            "class": "card-body",
            "style": css(**div_style_args),
        },
        {"class": "bslib-gap-spacing"} if fillable else None,
        *args,
        class_=class_,
        **kwargs,
    )

    if fill:
        tag = as_fill_item(tag)
    if fillable:
        tag = as_fillable_container(tag)

    return CardItem(tag)


############################################################################
# Full implementations
############################################################################


class CardItem:
    """
    A wrapper around a :class:`~htmltools.Tag` object that represents the content of a
    card item (e.g., :func:`~shiny.ui.card_header` or
    :func:`~shiny.ui.card_footer`).

    This class is used to allow for consecutive non-card items to be bundled into a
    single group within :func:`~shiny.ui.card`.

    Parameters
    ----------
    item
        A :class:`~htmltools.Tag` object that represents the content of a card item
        (e.g., :func:`~shiny.ui.card_header` or
        :func:`~shiny.ui.card_footer`).

    See Also
    --------
    * :func:`~shiny.ui.card` for creating a card component.
    * :func:`~shiny.ui.card_header` for creating a header within a card.
    * :func:`~shiny.ui.card_footer` for creating a footer within a card.
    """

    def __init__(
        self,
        item: TagChild,
    ):
        self._item = item

    def resolve(self) -> TagChild:
        """
        Resolves an object with the `CardItem` class by returning the `item` provided at initialization.

        Returns
        -------
        :
            The `item` provided at initialization.
        """
        return self._item

    def tagify(self) -> TagList:
        """
        Tagify the `item`

        Returns
        -------
        :
            A tagified :class:`~htmltools.TagList` object.
        """
        return TagList(self.resolve()).tagify()


@add_example()
def card_header(
    *args: TagChild | TagAttrs,
    container: TagFunction = tags.div,
    **kwargs: TagAttrValue,
) -> CardItem:
    """
    Card header container

    A general container for the "header" of a :func:`~shiny.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.ui.card`.

    The header has a different background color and border than the rest of the card.

    Parameters
    ----------
    *args
        Contents to the header container. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    container
        Method for the returned Tag object. Defaults to :func:`~shiny.ui.tags.div`.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A :class:`~shiny.ui.CardItem` object.

    See Also
    --------
    * :func:`~shiny.ui.card` for creating a card component.
    * :func:`~shiny.ui.card_footer` for creating a footer within the card.
    """
    return CardItem(
        container({"class": "card-header"}, *args, **kwargs),
    )


@add_example()
def card_footer(
    *args: TagChild | TagAttrs,
    **kwargs: TagAttrValue,
) -> CardItem:
    """
    Card footer container

    A general container for the "footer" of a :func:`~shiny.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.ui.card`.

    The footer has a different background color and border than the rest of the card.

    Parameters
    ----------
    *args
        Contents to the footer container. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    **kwargs
        Additional HTML attributes for the returned Tag.

    Returns
    -------
    :
        A :class:`~shiny.ui.CardItem` object.

    See Also
    --------
    * :func:`~shiny.ui.card` for creating a card component.
    * :func:`~shiny.ui.card_footer` for creating a footer within the card.
    """
    return CardItem(
        tags.div({"class": "card-footer"}, *args, **kwargs),
    )



import time

from shiny import App, Inputs, Outputs, Session, module, reactive, render, ui


@module.ui
def button_ui():
    return ui.TagList(
        ui.input_task_button("btn", label="Go"),
        ui.output_text("text_counter"),
    )


@module.server
def button_server(input: Inputs, output: Outputs, session: Session):
    counter = reactive.value(0)

    @render.text
    def text_counter():
        return f"Button clicked {counter()} times"

    @reactive.effect
    @reactive.event(input.btn)
    def increment_counter():
        time.sleep(0.5)
        counter.set(counter() + 1)


app_ui = ui.page_fluid(button_ui("mod1"))


def server(input: Inputs, output: Outputs, session: Session):
    button_server("mod1")


app = App(app_ui, server)



from typing import cast

import pandas as pd
from palmerpenguins import load_penguins  # pyright: ignore[reportMissingTypeStubs]

from shiny import reactive
from shiny.express import input, render, ui

penguins = cast(pd.DataFrame, load_penguins())

ui.input_action_button("update_filters", "Update filters")
ui.input_action_button("reset_filters", "Reset filters")

ui.h5("Current filters: ", {"class": "pt-2"})


@render.code
def penguins_code():
    return str(penguins_df.filter())


@render.data_frame
def penguins_df():
    return render.DataGrid(penguins, filters=True)


@reactive.effect
@reactive.event(input.update_filters)
async def _():
    await penguins_df.update_filter(
        [
            {"col": 0, "value": "Gentoo"},
            {"col": 2, "value": (50, None)},
            {"col": 3, "value": (None, 17)},
            {"col": 4, "value": (220, 225)},
        ],
    )


@reactive.effect
@reactive.event(input.reset_filters)
async def _():
    await penguins_df.update_filter(None)  # <<



from shiny import reactive
from shiny.express import input, ui

ui.input_action_button("show", "Show modal dialog")


@reactive.effect
@reactive.event(input.show)
def show_important_message():
    m = ui.modal(
        "This is a somewhat important message.",
        easy_close=True,
        footer=None,
    )
    ui.modal_show(m)
